import os

import setuptools
from setuptools.command.install import install


# Copied from setuptools:
# (https://github.com/pypa/setuptools/blob/7c859e017368360ba66c8cc591279d8964c031bc/setup.py#L40C6-L82)
class install_with_pth(install):
    """Custom install command to install a .pth file for distutils patching.

    This hack is necessary because there's no standard way to install behavior
    on startup (and it's debatable if there should be one). This hack (ab)uses
    the `extra_path` behavior in Setuptools to install a `.pth` file with
    implicit behavior on startup to give higher precedence to the local version
    of `distutils` over the version from the standard library.

    Please do not replicate this behavior.
    """

    _pth_name = "briefcase_debugger"
    _pth_contents = (
        "import briefcase_debugger; briefcase_debugger.start_remote_debugger()"
    )

    def initialize_options(self):
        install.initialize_options(self)
        self.extra_path = self._pth_name, self._pth_contents

    def finalize_options(self):
        install.finalize_options(self)
        self._restore_install_lib()

    def _restore_install_lib(self):
        """Undo secondary effect of `extra_path` adding to `install_lib`"""
        suffix = os.path.relpath(self.install_lib, self.install_libbase)

        if suffix.strip() == self._pth_contents.strip():
            self.install_lib = self.install_libbase


setuptools.setup(
    cmdclass={"install": install_with_pth},
)



import sys

import pytest

from briefcase.platforms.macOS import utils


@pytest.mark.skipif(sys.platform != "darwin", reason="Test runs only on macOS")
@pytest.mark.parametrize(
    ("uti", "result"),
    [
        (None, False),
        ("com.unknown.data", False),
        ("public.data", True),
        ("public.content", True),
        ("com.adobe.pdf", True),
    ],
)
def test_is_uti_core_type(uti, result):
    """Check if a UTI is a core type."""
    assert utils.is_uti_core_type(uti) == result


@pytest.mark.skipif(sys.platform != "darwin", reason="Test runs only on macOS")
def test_is_uti_core_type_with_nonexisting_coretypes_file(monkeypatch):
    """Test that is_uti_core_type returns None if the coretypes file doesn't exist."""
    monkeypatch.setattr(utils, "CORETYPES_PATH", "/does/not/exist")
    assert utils.is_uti_core_type("com.adobe.pdf") is False



from subprocess import CalledProcessError
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import (
    BriefcaseCommandError,
    MissingToolError,
    NetworkFailure,
    UnsupportedHostError,
)
from briefcase.integrations.wix import WiX

from ...utils import assert_url_resolvable
from .conftest import WIX_DOWNLOAD_URL, WIX_EXE_PATH


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.wix = "tool"

    tool = WiX.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.wix


@pytest.mark.parametrize("host_os", ["Darwin", "Linux", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{WiX.name} is not supported on {host_os}",
    ):
        WiX.verify(mock_tools)


def test_existing_wix_install(mock_tools, wix_path):
    """If there's an existing managed WiX install, the validator succeeds."""
    # Create a mock of a previously installed WiX version.
    wix_exe = wix_path / WIX_EXE_PATH
    wix_exe.parent.mkdir(parents=True)
    wix_exe.touch()

    mock_tools.subprocess.check_output.return_value = "5.0.2+aa65968c"

    wix = WiX.verify(mock_tools)

    # Version was checked
    mock_tools.subprocess.check_output.assert_called_once_with([wix_exe, "--version"])

    # No download was attempted
    assert mock_tools.file.download.call_count == 0

    # The returned paths are as expected
    assert wix.wix_exe == wix_exe


def test_download_missing(capsys, mock_tools, wix_path):
    """If there's no existing managed WiX install, it is downloaded and unpacked."""
    assert_download(mock_tools, wix_path)
    assert "WiX toolset was not found" in capsys.readouterr().out

    # Version was not checked
    mock_tools.subprocess.check_output.assert_not_called()

    # The WiX URL is resolvable
    assert_url_resolvable(WiX.verify(mock_tools).download_url)


@pytest.mark.parametrize(
    "exc",
    [PermissionError(), CalledProcessError(1, "wix")],
)
def test_download_unusable(capsys, mock_tools, wix_path, exc):
    """If the existing managed WiX install is unusable, it is reinstalled."""
    wix_exe = wix_path / WIX_EXE_PATH
    wix_exe.parent.mkdir(parents=True)
    wix_exe.touch()

    mock_tools.subprocess.check_output.side_effect = exc

    assert_download(mock_tools, wix_path)
    assert (
        f"WiX toolset is unusable ({type(exc).__name__}: {exc})"
        in capsys.readouterr().out
    )

    # Version was checked
    mock_tools.subprocess.check_output.assert_called_once_with([wix_exe, "--version"])


def test_download_version(capsys, mock_tools, wix_path):
    """If the existing managed WiX install is the wrong version, it is reinstalled."""
    wix_exe = wix_path / WIX_EXE_PATH
    wix_exe.parent.mkdir(parents=True)
    wix_exe.touch()

    mock_tools.subprocess.check_output.return_value = "5.0.1"

    assert_download(mock_tools, wix_path)
    assert "WiX toolset is an unsupported version (5.0.1)" in capsys.readouterr().out

    # Version was checked
    mock_tools.subprocess.check_output.assert_called_once_with([wix_exe, "--version"])


def assert_download(mock_tools, wix_path):
    # Mock the download
    wix_msi_path = wix_path.parent / "wix.msi"
    wix_msi_path.touch()
    mock_tools.file.download = MagicMock(return_value=wix_msi_path)

    # Verify the install. This will trigger a download
    wix = WiX.verify(mock_tools)

    # A download was initiated
    mock_tools.file.download.assert_called_with(
        url=WIX_DOWNLOAD_URL,
        download_path=wix_path.parent,
        role="WiX",
    )

    # The download was unpacked.
    mock_tools.subprocess.run.assert_called_with(
        ["msiexec", "/a", wix_msi_path, "/qn", f"TARGETDIR={wix_path}"], check=True
    )

    # The msi file was removed
    assert not wix_msi_path.exists()

    # The returned paths are as expected
    assert wix.wix_exe == wix_path / WIX_EXE_PATH


def test_dont_install_missing(mock_tools, tmp_path):
    """If there's no existing managed WiX install, and install is not requested, verify
    fails."""
    # Verify, but don't install. This will fail.
    with pytest.raises(MissingToolError):
        WiX.verify(mock_tools, install=False)

    # No download was initiated
    mock_tools.file.download.assert_not_called()


def test_dont_install_version(mock_tools, wix_path):
    """If the existing managed WiX install is the wrong version, and install is not
    requested, verify fails."""
    wix_exe = wix_path / WIX_EXE_PATH
    wix_exe.parent.mkdir(parents=True)
    wix_exe.touch()

    mock_tools.subprocess.check_output.return_value = "5.0.1"

    # Verify, but don't install. This will fail.
    with pytest.raises(MissingToolError):
        WiX.verify(mock_tools, install=False)

    # No download was initiated
    mock_tools.file.download.assert_not_called()


def test_download_fail(mock_tools, tmp_path):
    """If the download doesn't complete, the validator fails."""
    # Mock the download failure
    mock_tools.file.download = MagicMock(side_effect=NetworkFailure("mock"))

    # Verify the install. This will trigger a download
    with pytest.raises(NetworkFailure, match="Unable to mock"):
        WiX.verify(mock_tools)

    # A download was initiated
    mock_tools.file.download.assert_called_with(
        url=WIX_DOWNLOAD_URL,
        download_path=tmp_path / "tools",
        role="WiX",
    )

    # ... but the unpack didn't happen
    mock_tools.subprocess.run.assert_not_called()


def test_unpack_fail(capsys, mock_tools, wix_path):
    """If the download archive is corrupted, the validator fails."""
    # Mock the download
    wix_msi_path = wix_path.parent / "wix.msi"
    wix_msi_path.touch()
    mock_tools.file.download = MagicMock(return_value=wix_msi_path)

    # Mock an msiexec failure
    mock_tools.subprocess.run.side_effect = CalledProcessError(1, "msiexec")

    # Verify the install. This will trigger a download,
    # but the unpack will fail
    with pytest.raises(BriefcaseCommandError, match="interrupted or corrupted"):
        WiX.verify(mock_tools)

    # A download was initiated
    mock_tools.file.download.assert_called_with(
        url=WIX_DOWNLOAD_URL,
        download_path=wix_path.parent,
        role="WiX",
    )

    # The download was unpacked.
    mock_tools.subprocess.run.assert_called_with(
        ["msiexec", "/a", wix_msi_path, "/qn", f"TARGETDIR={wix_path}"], check=True
    )

    # The zip file was not removed
    assert wix_msi_path.exists()



import subprocess
from contextlib import suppress
from pathlib import Path

from briefcase.commands import (
    BuildCommand,
    DevCommand,
    OpenCommand,
    PublishCommand,
    UpdateCommand,
)
from briefcase.config import BaseConfig
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.rcedit import RCEdit
from briefcase.integrations.windows_sdk import WindowsSDK
from briefcase.platforms.windows import (
    WindowsCreateCommand,
    WindowsMixin,
    WindowsPackageCommand,
    WindowsRunCommand,
)


class WindowsAppMixin(WindowsMixin):
    output_format = "app"
    packaging_root = Path("src")
    supports_external_packaging = True

    def project_path(self, app):
        return self.bundle_path(app)


class WindowsAppCreateCommand(WindowsAppMixin, WindowsCreateCommand):
    description = "Create and populate a Windows app."


class WindowsAppUpdateCommand(WindowsAppCreateCommand, UpdateCommand):
    description = "Update an existing Windows app."
    supports_debugger = True


class WindowsAppOpenCommand(WindowsAppMixin, OpenCommand):
    description = "Open the folder containing an existing Windows app."


class WindowsAppBuildCommand(WindowsAppMixin, BuildCommand):
    description = "Build a Windows app."
    supports_debugger = True

    def verify_tools(self):
        super().verify_tools()
        RCEdit.verify(tools=self.tools)
        # The Windows SDK is only needed if it has previously been used to sign
        # the binary and MSI; therefore, ignore if it isn't available since the
        # stub app should not have any signatures to remove in that case.
        with suppress(BriefcaseCommandError):
            WindowsSDK.verify(tools=self.tools)

    def build_app(self, app: BaseConfig, **kwargs):
        """Build the application.

        :param app: The config object for the app
        """
        self.console.info("Building App...", prefix=app.app_name)

        # Move the stub binary in to the final executable location
        unbuilt_binary_path = self.unbuilt_executable_path(app)

        # We can run this test on non-Windows platforms, but when we run on
        # non-windows platforms, the ".exe" suffix doesn't exist, so it doesn't
        # hit this branch. That's not actually a problem, as long as we *are*
        # hitting the branch under Windows.
        if unbuilt_binary_path.exists():  # pragma: no-cover-if-not-windows
            with self.console.wait_bar("Renaming stub binary..."):
                unbuilt_binary_path.rename(self.binary_executable_path(app))

        if hasattr(self.tools, "windows_sdk"):
            # If an app has been packaged and code signed previously, then the digital
            # signature on the app binary must be removed before re-building the app.
            # It is not safe to use RCEdit on signed binaries since it corrupts them.
            with self.console.wait_bar(
                "Removing any digital signatures from stub app..."
            ):
                try:
                    self.tools.subprocess.check_output(
                        [
                            self.tools.windows_sdk.signtool_exe,
                            "remove",
                            "-s",
                            self.binary_path(app).relative_to(self.bundle_path(app)),
                        ],
                        cwd=self.bundle_path(app),
                        quiet=1,
                    )
                except subprocess.CalledProcessError as e:
                    # Ignore this error from signtool since it is logged if the file
                    # is not currently signed
                    if "error: 0x00000057" not in e.stdout:
                        self.tools.subprocess.output_error(e)
                        raise BriefcaseCommandError(f"""\
Failed to remove any existing digital signatures from the stub app.

Recreating the app layout may also help resolve this issue:

    $ briefcase create {self.platform} {self.output_format}

""") from e

        with self.console.wait_bar("Setting stub app details..."):
            try:
                self.tools.subprocess.run(
                    [
                        self.tools.rcedit.rcedit_path,
                        self.binary_path(app).relative_to(self.bundle_path(app)),
                        "--set-version-string",
                        "CompanyName",
                        app.author,
                        # "FileDescription" is the label that appears as a grouping
                        # in the Task Manager when the application runs.
                        "--set-version-string",
                        "FileDescription",
                        app.formal_name,
                        "--set-version-string",
                        "FileVersion",
                        str(app.version),
                        "--set-version-string",
                        "InternalName",
                        app.module_name,
                        "--set-version-string",
                        "OriginalFilename",
                        self.binary_path(app).name,
                        "--set-version-string",
                        "ProductName",
                        app.formal_name,
                        "--set-version-string",
                        "ProductVersion",
                        str(app.version),
                        "--set-icon",
                        "icon.ico",
                    ],
                    check=True,
                    cwd=self.bundle_path(app),
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(f"""\
Unable to update details on stub app for {app.app_name}.

This may be caused by a virus scanner misidentifying the Briefcase build as malicious
activity. Try disabling your virus checker, and re-run briefcase build.
""") from e


class WindowsAppRunCommand(WindowsAppMixin, WindowsRunCommand):
    description = "Run a Windows app."


class WindowsAppDevCommand(WindowsAppMixin, DevCommand):
    description = "Run a Windows app in development mode."


class WindowsAppPackageCommand(WindowsAppMixin, WindowsPackageCommand):
    description = "Package a Windows app."


class WindowsAppPublishCommand(WindowsAppMixin, PublishCommand):
    description = "Publish a Windows app."


# Declare the briefcase command bindings
create = WindowsAppCreateCommand
update = WindowsAppUpdateCommand
open = WindowsAppOpenCommand
build = WindowsAppBuildCommand
run = WindowsAppRunCommand
package = WindowsAppPackageCommand
publish = WindowsAppPublishCommand
dev = WindowsAppDevCommand



import errno
import re
import subprocess
import sys
import webbrowser
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path, PurePosixPath
from textwrap import dedent, indent
from typing import Any
from zipfile import ZipFile

from briefcase.console import Console
from briefcase.exceptions import (
    BriefcaseCommandError,
    BriefcaseConfigError,
    UnsupportedCommandError,
)

if sys.version_info >= (3, 11):  # pragma: no-cover-if-lt-py311
    import tomllib
else:  # pragma: no-cover-if-gte-py311
    import tomli as tomllib

import tomli_w

from briefcase.commands import (
    BuildCommand,
    CreateCommand,
    DevCommand,
    OpenCommand,
    PackageCommand,
    PublishCommand,
    RunCommand,
    UpdateCommand,
)
from briefcase.config import AppConfig

# Banner templates (Constants used in write_inserts)
HTML_BANNER = (
    "<!--------------------------------------------------\n"
    " * {package}\n"
    " -------------------------------------------------->\n"
    "{content}"
)

CSS_BANNER = (
    "/**************************************************\n"
    " * {package}\n"
    " *************************************************/\n"
    "{content}"
)


class StaticWebMixin:
    output_format = "static"
    platform = "web"
    platform_target_version = "0.3.21"

    def project_path(self, app):
        return self.bundle_path(app) / "www"

    def binary_path(self, app):
        return self.bundle_path(app) / "www/index.html"

    def static_path(self, app):
        return self.project_path(app) / "static"

    def wheel_path(self, app):
        return self.static_path(app) / "wheels"

    def distribution_path(self, app):
        return self.dist_path / f"{app.formal_name}-{app.version}.web.zip"


class StaticWebCreateCommand(StaticWebMixin, CreateCommand):
    description = "Create and populate a static web project."


class StaticWebUpdateCommand(StaticWebCreateCommand, UpdateCommand):
    description = "Update an existing static web project."


class StaticWebOpenCommand(StaticWebMixin, OpenCommand):
    description = "Open the folder containing an existing static web project."


class StaticWebBuildCommand(StaticWebMixin, BuildCommand):
    description = "Build a static web project."

    def write_inserts(
        self,
        app: AppConfig,
        filename: Path,
        inserts: dict[str, dict[str, str]],
    ):
        """Write inserts into an existing file.

        This function looks for start and end markers in the named file and
        replaces the content inside the markers with the inserted content.

        Multiple formats of insert marker are inspected to accommodate HTML
        and CSS/JS comment conventions:
        * HTML: `<!--@@ insert:start @@--> and <!--@@ insert:end @@-->`
        * CSS/JS: `/*@@ insert:start @@*/ and /*@@ insert:end @@*/`

        Inserts and package contributions are processed in sorted order to ensure
        deterministic builds.

        :param app: The application whose `pyscript.toml` is being written.
        :param filename: The file whose insert is to be written.
        :param inserts: The inserts for the file. A 2 level dictionary, keyed by
            the name of the insert to add, and then package that contributed the
            insert.
        """
        # Load file content, skip if file not found
        target_path = self.project_path(app) / filename
        try:
            file_text = target_path.read_text(encoding="utf-8")
        except FileNotFoundError:
            self.console.warning(f"  Target {filename} not found; skipping inserts.")
            return

        # Each insert slot and its package contributions are processed in sorted order
        for insert, pkg_contribs in sorted(inserts.items()):
            # Build bodies from the same contributions
            html_body = "\n".join(
                HTML_BANNER.format(package=pkg, content=text)
                for pkg, text in sorted(pkg_contribs.items())
                if text
            )
            css_body = "\n".join(
                CSS_BANNER.format(package=pkg, content=text)
                for pkg, text in sorted(pkg_contribs.items())
                if text
            )
            body_map = {"html": html_body, "css": css_body}

            slot = re.escape(insert)

            # Build the compiled patterns directly once per slot
            html_marker = (
                re.compile(
                    rf"(^[ \t]*)<!--@@ {slot}:start @@-->.*?<!--@@ {slot}:end @@-->",
                    flags=re.MULTILINE | re.DOTALL,
                ),
                (
                    r"{indent}<!--@@ {insert}:start @@-->\n"
                    r"{content}{indent}<!--@@ {insert}:end @@-->"
                ),
                "html",
            )
            css_marker = (
                re.compile(
                    rf"(^[ \t]*)/\*@@ {slot}:start @@\*/.*?/\*@@ {slot}:end @@\*/",
                    flags=re.MULTILINE | re.DOTALL,
                ),
                (
                    r"{indent}/*@@ {insert}:start @@*/\n"
                    r"{content}{indent}/*@@ {insert}:end @@*/"
                ),
                "css",
            )

            # Apply all matching marker styles
            any_match = False
            for pattern, repl_tmpl, kind in [html_marker, css_marker]:
                # Search for pattern within file.
                pattern_found = pattern.search(file_text)
                if pattern_found:
                    # Indent content to align with markers.
                    indented_content = indent(
                        body_map.get(kind, "").strip("\n") + "\n",
                        pattern_found.group(1),
                    )
                    file_text = pattern.sub(
                        repl_tmpl.format(
                            indent=pattern_found.group(1),
                            insert=insert,
                            content=indented_content,
                        ),
                        file_text,
                    )
                    any_match = True

            if not any_match:
                self.console.warning(
                    f"  Slot '{insert}' markers not found in {filename}; skipping."
                )

        # Save modified content
        target_path.write_text(file_text, encoding="utf-8")

    def _append_pyscript_insert(
        self,
        app: AppConfig,
        filename: Path,
        pyscript_version: str,
        inserts: dict[str, dict[str, dict[str, str]]],
    ):
        """Write pyscript version into an existing html file.

        This function creates an insert for PyScript.

        :param app: The application whose html file is being written.
        :param filename: The html file whose pyscript version is to be written.
        :param pyscript_version: The pyscript version number to be inserted.
        :param inserts: Nested dict of inserts keyed by target - insert - package.
        """
        package_key = "briefcase"
        target = filename
        head_insert = "head-python"
        body_insert = "body-python"

        # PyScript definitions for insertion:
        head_content = dedent(f"""\
            <script type="module">
                // Hide the splash screen when the page is ready.
                import {{ hooks }} from "https://pyscript.net/releases/{pyscript_version}/core.js";
                hooks.main.onReady.add(() => {{
                    document.getElementById("briefcase-splash").classList.add("hidden");
                }});
            </script>

            <link rel="stylesheet" href="https://pyscript.net/releases/{pyscript_version}/core.css">
            <script type="module" src="https://pyscript.net/releases/{pyscript_version}/core.js"></script>
            """)
        body_content = dedent(f"""\
            <script type="py" async="false" config="pyscript.toml">
                import runpy
                result = runpy.run_module(
                    "{app.app_name}", run_name="__main__", alter_sys=True
                )
            </script>
            """)

        pkg_map_head = inserts.setdefault(target, {}).setdefault(head_insert, {})
        pkg_map_head[package_key] = head_content

        pkg_map_body = inserts.setdefault(target, {}).setdefault(body_insert, {})
        pkg_map_body[package_key] = body_content

    def _handle_legacy_css(
        self,
        wheel: ZipFile,
        path: PurePosixPath,
        package_key: str,
        inserts: dict[str, dict[str, dict[str, str]]],
    ) -> None:
        """Handle legacy CSS under /static/*.css and add to style.css.

        Emits a deprecation warning for every legacy CSS file discovered.

        :param wheel: Open wheel ZipFile being processed.
        :param path: PurePosixPath object of the file inside the wheel.
        :param filename: Filename string inside the wheel.
        :param package_key: Provenance label (e.g. "name version").
        :param inserts: Nested dict of inserts keyed by target - insert - package.
        """
        # Warn on every legacy usage
        self.console.warning(
            f"    {Path(wheel.filename).name}: legacy '/static' CSS file {path} "
            "detected.\n"
            "     Static file handling has been deprecated; this file should be "
            "converted into an insert."
        )

        css_text = wheel.read(str(path)).decode("utf-8")

        rel_inside = "/".join(path.parts[2:])
        contrib_key = f"{package_key} (legacy static CSS: {rel_inside})"

        # Add CSS content to style.css insert slot
        target = "static/css/style.css"
        insert = "css"
        pkg_map = inserts.setdefault(target, {}).setdefault(insert, {})
        if pkg_map.get(contrib_key):
            pkg_map[contrib_key] += "\n" + css_text
        else:
            pkg_map[contrib_key] = css_text

    def _handle_insert(
        self,
        wheel: ZipFile,
        path: PurePosixPath,
        package_key: str,
        inserts: dict[str, dict[str, dict[str, str]]],
    ) -> None:
        """Handle deploy/inserts/<target>~<insert> files and register inserts.

        :param wheel: Open wheel ZipFile being processed.
        :param path: PurePosixPath object of the file inside the wheel.
        :param package_key: Provenance label (e.g. "name version").
        :param inserts: Nested dict of inserts keyed by target - insert - package.
        """
        self.console.info(f"    Found {path}")

        rel_inside = "/".join(path.parts[3:])

        if not rel_inside or rel_inside.endswith("/"):
            self.console.warning(
                f"    {path}: skipping; "
                "not a valid insert file (empty path or directory)."
            )
            return

        if "~" not in rel_inside:
            self.console.warning(
                f"    {path}: skipping; filename must match '<target>~<insert>'."
            )
            return

        # Preserve any '~' that might exist in the target path
        target, insert = rel_inside.rsplit("~", 1)
        self.console.info(f"    {path}: Adding {insert} insert for {target}")

        text = wheel.read(str(path)).decode("utf-8")

        contrib_key = f"{package_key} (deploy insert: {rel_inside} from {path})"
        pkg_map = inserts.setdefault(target, {}).setdefault(insert, {})
        if pkg_map.get(contrib_key):
            pkg_map[contrib_key] += "\n" + text
        else:
            pkg_map[contrib_key] = text

    def _process_wheel(
        self,
        wheelfile,
        inserts: dict[str, dict[str, dict[str, str]]],
    ) -> None:
        """Process a wheel to collect insert and style content.

        Scans the wheel for:
        * Legacy CSS - .css files under /static/ added to style.css.
        * Deploy inserts - files under /deploy/inserts/<target>~<insert>.

        :param wheelfile: Path to the wheel file.
        :param inserts: Nested dict of inserts keyed by target - insert - package.
        """
        name_parts = wheelfile.name.split("-")
        package_key = f"{name_parts[0]} {name_parts[1]}"

        with ZipFile(wheelfile) as wheel:
            for filename in sorted(wheel.namelist()):
                # Skip directories and shallow paths
                path = PurePosixPath(filename)
                parts = path.parts

                # Legacy CSS handling
                if (
                    len(parts) > 1
                    and parts[1] == "static"
                    and path.suffix.lower() == ".css"
                ):
                    self._handle_legacy_css(wheel, path, package_key, inserts)

                # New deploy/inserts handling
                elif len(parts) >= 3 and parts[1] == "deploy" and parts[2] == "inserts":
                    self._handle_insert(wheel, path, package_key, inserts)

    def extract_pyscript_config(self, wheels):
        """Process multiple wheels to gather a config.toml and a base pyscript.toml
        file.

        :param wheels: A list of wheel files to be scanned.
        """
        config_package = None
        config_package_list = []
        config_filename = None
        pyscript_version = "2024.11.1"
        pyscript_config = None

        # Find packages containing a config.toml file.
        for wheelfile in wheels:
            with ZipFile(wheelfile) as wheel:
                for filename in wheel.namelist():
                    path = Path(filename)
                    if (
                        len(path.parts) == 3
                        and path.parts[1] == "deploy"
                        and path.name == "config.toml"
                    ):
                        self.console.info(f"    Found {filename}")
                        config_package_list.append(wheelfile)
                        config_filename = filename

        # Return a blank pyscript config if no configuration file is found.
        if len(config_package_list) == 0:
            pyscript_config = {}
        # Raise an error if more than one configuration file is supplied.
        elif len(config_package_list) > 1:
            raise BriefcaseConfigError(
                "Only one deployment configuration file can be supplied. "
                f"Initial config.toml found in package: {config_package}; "
                f"Duplicate config.toml found in package: {wheel.filename}"
            )
        # Gather a deployment configuration file from the package. For now, this is a
        # pyscript.toml as no other backend is currently supported.
        else:
            with (
                ZipFile(config_package_list[0]) as wheel,
                wheel.open(config_filename) as config_file,
            ):
                config_data = tomllib.load(config_file)

                # Gather implementation from config.toml
                implementation = config_data.get("implementation", None)

                # Currently, only pyscript is supported.
                # Warn if another implementation is found,
                # but fail safe to using pyscript with a warning.
                if implementation is None:
                    self.console.warning(
                        "No web implementation specified. Defaulting to 'pyscript'."
                    )
                elif implementation != "pyscript":
                    self.console.warning(
                        "At present, 'pyscript' is the only supported web "
                        "implementation. This project may not work correctly."
                    )

                # Get pyscript version from config.toml. Use default if not present.
                pyscript_version = config_data.get("pyscript", {}).get(
                    "version", pyscript_version
                )

                # Extract pyscript.toml
                pyscript_path = config_filename.replace("config.toml", "pyscript.toml")
                try:
                    with wheel.open(pyscript_path) as pyscript_file:
                        pyscript_config = tomllib.load(pyscript_file)
                except tomllib.TOMLDecodeError as e:
                    raise BriefcaseConfigError(
                        f"pyscript.toml content isn't valid TOML: {e}"
                    ) from e
                except (KeyError, FileNotFoundError):
                    self.console.info(
                        "Pyscript configuration file not found in "
                        f"{config_package_list[0]}. Using default configuration."
                    )
                    pyscript_config = {}

        return pyscript_config, pyscript_version

    def build_app(self, app: AppConfig, **kwargs):
        """Build the static web deployment for the application.

        :param app: The application to build
        """
        self.console.info("Building web project...", prefix=app.app_name)

        if self.wheel_path(app).exists():
            with self.console.wait_bar("Removing old wheels..."):
                self.tools.shutil.rmtree(self.wheel_path(app))

        self.wheel_path(app).mkdir(parents=True)

        with self.console.wait_bar("Building app wheel..."):
            try:
                self.tools.subprocess.run(
                    [
                        sys.executable,
                        "-u",
                        "-X",
                        "utf8",
                        "-m",
                        "wheel",
                        "pack",
                        self.app_path(app),
                        "--dest-dir",
                        self.wheel_path(app),
                    ],
                    check=True,
                    encoding="UTF-8",
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to build wheel for app {app.app_name!r}"
                ) from e

        with self.console.wait_bar("Installing wheels for requirements..."):
            try:
                self.tools.subprocess.run(
                    [
                        sys.executable,
                        "-u",
                        "-X",
                        "utf8",
                        "-m",
                        "pip",
                        "wheel",
                        "--wheel-dir",
                        self.wheel_path(app),
                        "-r",
                        self.bundle_path(app) / "requirements.txt",
                    ]
                    + (["-vv"] if self.console.is_deep_debug else []),
                    check=True,
                    encoding="UTF-8",
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to install requirements for app {app.app_name!r}"
                ) from e

        with self.console.wait_bar("Writing Pyscript configuration file..."):
            # Load any pre-existing pyscript.toml provided by a GUI Toolkit. If the file
            # doesn't exist, assume an empty pyscript.toml as a starting point.
            config, pyscript_version = self.extract_pyscript_config(
                self.wheel_path(app).glob("*.whl")
            )

            # Add the packages declaration to the existing pyscript.toml.
            # Ensure that we're using Unix path separators, as the content
            # will be parsed by pyscript in the browser.
            config["packages"] = [
                f"/{'/'.join(wheel.relative_to(self.project_path(app)).parts)}"
                for wheel in sorted(self.wheel_path(app).glob("*.whl"))
            ]

            # Parse any additional pyscript.toml content, and merge it into
            # the overall content
            try:
                extra = tomllib.loads(app.extra_pyscript_toml_content)
                config.update(extra)
            except tomllib.TOMLDecodeError as e:
                raise BriefcaseConfigError(
                    f"Extra pyscript.toml content isn't valid TOML: {e}"
                ) from e
            except AttributeError:
                pass

            # Write the final configuration
            with (self.project_path(app) / "pyscript.toml").open("wb") as f:
                tomli_w.dump(config, f)

        with self.console.wait_bar("Compiling static web content from wheels..."):
            inserts: dict[str, dict[str, dict[str, str]]] = {}

            for wheelfile in sorted(self.wheel_path(app).glob("*.whl")):
                self.console.info(f"  Processing {wheelfile.name}...")
                self._process_wheel(
                    wheelfile=wheelfile,
                    inserts=inserts,
                )

            # Add pyscript insertion to inserts
            self._append_pyscript_insert(app, "index.html", pyscript_version, inserts)

            # Write inserts per target
            for target, target_inserts in sorted(inserts.items()):
                self.write_inserts(app, Path(target), target_inserts)

        return {}


class HTTPHandler(SimpleHTTPRequestHandler):
    """Convert any HTTP request into a path request on the static content folder."""

    server: "LocalHTTPServer"

    def translate_path(self, path):
        return str(self.server.base_path / path[1:])

    def end_headers(self):
        self.send_header("Cache-Control", "no-cache, no-store, must-revalidate")
        self.send_header("Pragma", "no-cache")
        self.send_header("Expires", "0")
        super().end_headers()

    # Strip out ANSI escape sequences from messages. This was added in security releases
    # before Python 3.12.
    _control_char_table = str.maketrans(
        {c: rf"\x{c:02x}" for c in [*range(0x20), *range(0x7F, 0xA0)]}
    )
    _control_char_table[ord("\\")] = r"\\"

    def log_message(self, format: str, *args: Any) -> None:
        message = (format % args).translate(self._control_char_table)
        self.server.logger.info(
            f"{self.address_string()} - - [{self.log_date_time_string()}] {message}"
        )


class LocalHTTPServer(ThreadingHTTPServer):
    """An HTTP server that serves local static content."""

    def __init__(
        self,
        base_path,
        host,
        port,
        RequestHandlerClass=HTTPHandler,
        *,
        logger: Console,
    ):
        self.base_path = base_path
        self.logger = logger
        super().__init__((host, port), RequestHandlerClass)


class StaticWebRunCommand(StaticWebMixin, RunCommand):
    description = "Run a static web project."

    def add_options(self, parser):
        super().add_options(parser)
        parser.add_argument(
            "--host",
            default="localhost",
            help="The host on which to run the server (default: localhost)",
            required=False,
        )
        parser.add_argument(
            "-p",
            "--port",
            default=8080,
            type=int,
            help="The port on which to run the server (default: 8080)",
            required=False,
        )
        parser.add_argument(
            "--no-browser",
            action="store_false",
            dest="open_browser",
            help="Don't open a web browser on the newly opened server",
            required=False,
        )

    def run_app(
        self,
        app: AppConfig,
        passthrough: list[str],
        host,
        port,
        open_browser,
        **kwargs,
    ):
        """Start the application.

        :param app: The config object for the app
        :param passthrough: The list of arguments to pass to the app
        :param host: The host on which to run the server
        :param port: The port on which to run the server
        :param open_browser: Should a browser be opened on the newly started server.
        """
        if app.test_mode:
            raise BriefcaseCommandError("Briefcase can't run web apps in test mode.")

        self.console.info("Starting web server...", prefix=app.app_name)

        # At least for now, there's no easy way to pass arguments to a web app.
        if passthrough:
            self.console.warning(f"Ignoring passthrough arguments: {passthrough}")

        httpd = None
        try:
            # Create a local HTTP server
            try:
                httpd = LocalHTTPServer(
                    self.project_path(app),
                    host=host,
                    port=port,
                    logger=self.console,
                )
            except OSError as e:
                if e.errno in (errno.EADDRINUSE, errno.ENOSR):
                    self.console.warning(
                        f"Using a system-allocated port since port {port} is already "
                        "in use. Use -p/--port to manually specify a port."
                    )
                    httpd = LocalHTTPServer(
                        self.project_path(app),
                        host=host,
                        port=0,
                        logger=self.console,
                    )
                else:
                    raise

            # Extract the host and port from the server. This is needed
            # because specifying a port of 0 lets the server pick a port.
            host, port = httpd.socket.getsockname()
            url = f"http://{host}:{port}"

            self.console.info(f"Web server open on {url}")
            # If requested, open a browser tab on the newly opened server.
            if open_browser:
                webbrowser.open_new_tab(url)

            self.console.info(
                "Web server log output (type CTRL-C to stop log)...",
                prefix=app.app_name,
            )
            self.console.info("=" * 75)

            # Run the server.
            httpd.serve_forever()
        except PermissionError as e:
            if port < 1024:
                raise BriefcaseCommandError(
                    "Unable to start web server; Permission denied. "
                    "Try using a port > 1023."
                ) from e
            else:
                raise BriefcaseCommandError(
                    "Unable to start web server; Permission denied. "
                    "Did you specify a valid host and port?"
                ) from e
        except OSError as e:
            if e.errno in (errno.EADDRNOTAVAIL, errno.ENOSTR):
                raise BriefcaseCommandError(
                    f"Unable to start web server. {host} is not a valid hostname."
                ) from e
            else:
                raise BriefcaseCommandError(f"Unable to start web server. {e}") from e
        except OverflowError as e:
            raise BriefcaseCommandError(
                "Unable to start web server. Port must be in the range 0-65535."
            ) from e
        except KeyboardInterrupt:
            # CTRL-C is the accepted way to stop the server.
            httpd.shutdown()
        finally:
            if httpd:
                with self.console.wait_bar("Shutting down server..."):
                    httpd.server_close()

            # Not sure why, but this is needed to mollify coverage for the
            # "test_cleanup_server_error" case. Without this pass, a missing branch
            # is reported for the "if httpd: -> exit" branch
            pass  # noqa: PIE790 (unnecessary pass)

        return {}


class StaticWebPackageCommand(StaticWebMixin, PackageCommand):
    description = "Package a static web app."

    @property
    def packaging_formats(self):
        return ["zip"]

    @property
    def default_packaging_format(self):
        return "zip"

    def package_app(self, app: AppConfig, **kwargs):
        """Package an app for distribution.

        :param app: The app to package.
        """
        self.console.info(
            "Packaging web app for distribution...",
            prefix=app.app_name,
        )

        with self.console.wait_bar("Building archive..."):
            self.tools.shutil.make_archive(
                self.distribution_path(app).with_suffix(""),
                format="zip",
                root_dir=self.project_path(app),
            )


class StaticWebPublishCommand(StaticWebMixin, PublishCommand):
    description = "Publish a static web app."


class StaticWebDevCommand(StaticWebMixin, DevCommand):
    description = "Run a static web project in development mode. (Work in progress)"

    @property
    def venv_name(self) -> str:
        """Returns the name of the virtual environment directory.

        The web environment uses a different venv because a web deployment has
        different requirements to a desktop deployment. The name is the same as the
        desktop deployment, but with `.web` appended.

        :returns: Name for virtual environment directory
        """
        return f"{DevCommand.venv_name.fget(self)}.web"

    def run_dev_app(self, app: AppConfig, **options):
        raise UnsupportedCommandError(
            platform="web",
            output_format="static",
            command="dev",
        )


# Declare the briefcase command bindings
create = StaticWebCreateCommand
update = StaticWebUpdateCommand
open = StaticWebOpenCommand
build = StaticWebBuildCommand
run = StaticWebRunCommand
package = StaticWebPackageCommand
publish = StaticWebPublishCommand
dev = StaticWebDevCommand



import pytest

from briefcase.exceptions import UnsupportedHostError
from briefcase.platforms.web.static import StaticWebCreateCommand


@pytest.fixture
def create_command(dummy_console, tmp_path):
    return StaticWebCreateCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("host_os", ["WeirdOS"])
def test_unsupported_host_os(create_command, host_os):
    """Error raised for an unsupported OS."""
    create_command.tools.host_os = host_os

    with pytest.raises(UnsupportedHostError, match="This command is not supported on"):
        create_command()



import os

import pytest

from briefcase.console import Console

from ...utils import DummyConsole


@pytest.fixture
def raw_console() -> Console:
    console = Console()
    yield console
    console.close()


@pytest.fixture
def console(monkeypatch) -> DummyConsole:
    console = DummyConsole()
    # default console is always interactive
    monkeypatch.setattr(os, "isatty", lambda _: True)
    yield console
    console.close()


@pytest.fixture
def disabled_console() -> DummyConsole:
    console = DummyConsole(input_enabled=False)
    yield console
    console.close()


@pytest.fixture
def non_interactive_console(console, monkeypatch) -> DummyConsole:
    monkeypatch.setattr(os, "isatty", lambda _: False)
    return console



from __future__ import annotations

import enum
import json
import platform
from abc import ABC, abstractmethod
from importlib import metadata
from pathlib import Path
from typing import TYPE_CHECKING, TypedDict

import briefcase

if TYPE_CHECKING:
    # avoid circular imports
    from briefcase.commands.run import RunCommand
    from briefcase.config import AppConfig


def _is_editable_pep610(dist_name: str) -> bool:
    """Check if briefcase is installed as editable build.

    The check requires, that the tool that installs briefcase support PEP610 (eg. pip
    since v20.1).
    """
    try:
        dist = metadata.distribution(dist_name)
    except metadata.PackageNotFoundError:
        raise

    direct_url = dist.read_text("direct_url.json")
    if direct_url is None:
        return False

    try:
        data = json.loads(direct_url)
        return data.get("dir_info", {}).get("editable", False)
    except Exception:
        return False


IS_EDITABLE = _is_editable_pep610("briefcase")
REPO_ROOT = Path(__file__).parent.parent.parent.parent if IS_EDITABLE else None


def get_debugger_requirement(package_name: str, extras: str = ""):
    """Get the requirement of a debugger support package.

    On editable installs of briefcase the path to the local package is used, to simplify
    the development of the debugger support packages. On normal installs the local
    version is not available, so the package from pypi is used, that corresponds to the
    version of briefcase.

    :param package_name: The name of the debugger support package.
    :param extras: Optional extras to add to the package requirement. Including square
        brackets. E.g. "[debugpy]".
    :return: The package requirement.
    """
    if IS_EDITABLE and REPO_ROOT is not None:
        local_path = REPO_ROOT / "debugger"
        if local_path.exists() and local_path.is_dir():
            return f"{local_path}{extras}"
    return f"{package_name}{extras}=={briefcase.__version__}"


class AppPathMappings(TypedDict):
    device_sys_path_regex: str
    device_subfolders: list[str]
    host_folders: list[str]


class AppPackagesPathMappings(TypedDict):
    sys_path_regex: str
    host_folder: str


class DebuggerConfig(TypedDict):
    debugger: str
    host: str
    port: int
    host_os: str
    app_path_mappings: AppPathMappings | None
    app_packages_path_mappings: AppPackagesPathMappings | None


class DebuggerConnectionMode(str, enum.Enum):
    SERVER = "server"
    CLIENT = "client"


class BaseDebugger(ABC):
    """Definition for a plugin that defines a new Briefcase debugger."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the name debugger."""

    @property
    @abstractmethod
    def connection_mode(self) -> DebuggerConnectionMode:
        """Return the connection mode of the debugger."""

    @property
    @abstractmethod
    def debugger_support_pkg(self) -> str:
        """Get the name of the debugger support package."""

    def get_env_config(
        self,
        cmd: RunCommand,
        app: AppConfig,
    ) -> str:
        """Get the environment config to start the debugger.

        :param cmd: The command that starts the debugger
        :param app: The app to be debugged
        :returns: The remote debugger configuration
        """
        config = DebuggerConfig(
            debugger=app.debugger.name,
            host=app.debugger_host,
            port=app.debugger_port,
            host_os=platform.system(),
            app_path_mappings=cmd.debugger_app_path_mappings(app),
            app_packages_path_mappings=cmd.debugger_app_packages_path_mapping(app),
        )
        return json.dumps(config)



from unittest.mock import MagicMock


def test_run_passes_through_to_subprocess(noop_context, mock_noop_subprocess_setup):
    """Run passes all arguments to tools.subprocess.run."""
    mocks = mock_noop_subprocess_setup
    mock_result = MagicMock()
    mocks["subprocess"].run.return_value = mock_result

    args = ["python", "-m", "pip", "install", "package"]
    kwargs = {"check": True, "cwd": "/tmp", "env": {"PATH": "/usr/bin"}}

    result = noop_context.run(args, **kwargs)

    mocks["subprocess"].run.assert_called_once_with(args, **kwargs)
    assert result is mock_result


def test_run_with_no_kwargs(noop_context, mock_noop_subprocess_setup):
    """Run works with just args, no kwargs."""
    mocks = mock_noop_subprocess_setup
    mock_result = MagicMock()
    mocks["subprocess"].run.return_value = mock_result

    args = ["python", "--version"]
    result = noop_context.run(args)

    mocks["subprocess"].run.assert_called_once_with(args)
    assert result is mock_result


def test_run_preserves_original_kwargs(noop_context, mock_noop_subprocess_setup):
    """Run does not modify the original kwargs dict."""
    mocks = mock_noop_subprocess_setup
    mock_result = MagicMock()
    mocks["subprocess"].run.return_value = mock_result

    original_kwargs = {"check": True, "env": {"CUSTOM": "value"}}
    kwargs_copy = original_kwargs.copy()

    result = noop_context.run(["test"], **original_kwargs)

    assert original_kwargs == kwargs_copy
    mocks["subprocess"].run.assert_called_once_with(["test"], **original_kwargs)
    assert result is mock_result



import platform
import subprocess
from zipfile import ZipFile

import pytest

from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.skipif(
    platform.system() != "Darwin",
    reason="ditto tests can only be performed on macOS",
)
def test_ditto(
    package_command,
    first_app_with_binaries,
    tmp_path,
):
    """An app archive can be built with ditto."""
    app_path = (
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "macos"
        / "app"
        / "First App.app"
    )
    archive_path = tmp_path / "base_path/build/first-app/macos/app/archive.zip"

    # Reconnect the subprocess.run with its live implementation
    package_command.tools.subprocess.run.side_effect = subprocess.run

    # Build the ditto archive
    package_command.ditto_archive(app_path, archive_path)

    # The archive contains the app as the only top level element.
    with ZipFile(archive_path) as archive:
        # zip file can include a “__MACOSX” folder for each document that contains information about
        # the file useful for Finder and will not be in the unzipped set of files
        archived_files = [
            fn for fn in archive.namelist() if not fn.startswith("__MACOSX/")
        ]

        assert sorted(archived_files) == [
            "First App.app/",
            "First App.app/Contents/",
            "First App.app/Contents/Frameworks/",
            "First App.app/Contents/Frameworks/Extras.framework/",
            "First App.app/Contents/Frameworks/Extras.framework/Extras",
            "First App.app/Contents/Frameworks/Extras.framework/Resources/",
            "First App.app/Contents/Frameworks/Extras.framework/Resources/Info.plist",
            "First App.app/Contents/Frameworks/Extras.framework/Versions/",
            "First App.app/Contents/Frameworks/Extras.framework/Versions/1.2/",
            "First App.app/Contents/Frameworks/Extras.framework/Versions/1.2/Extras",
            "First App.app/Contents/Frameworks/Extras.framework/Versions/1.2/libs/",
            "First App.app/Contents/Frameworks/Extras.framework/Versions/1.2/libs/extras.dylib",
            "First App.app/Contents/Frameworks/Extras.framework/Versions/Current",
            "First App.app/Contents/Info.plist",
            "First App.app/Contents/MacOS/",
            "First App.app/Contents/MacOS/First App",
            "First App.app/Contents/Resources/",
            "First App.app/Contents/Resources/app_packages/",
            "First App.app/Contents/Resources/app_packages/Extras.app/",
            "First App.app/Contents/Resources/app_packages/Extras.app/Contents/",
            "First App.app/Contents/Resources/app_packages/Extras.app/Contents/MacOS/",
            "First App.app/Contents/Resources/app_packages/Extras.app/Contents/MacOS/Extras",
            "First App.app/Contents/Resources/app_packages/first.other",
            "First App.app/Contents/Resources/app_packages/first_dylib.dylib",
            "First App.app/Contents/Resources/app_packages/first_so.so",
            "First App.app/Contents/Resources/app_packages/other_binary",
            "First App.app/Contents/Resources/app_packages/second.other",
            "First App.app/Contents/Resources/app_packages/special.binary",
            "First App.app/Contents/Resources/app_packages/subfolder/",
            "First App.app/Contents/Resources/app_packages/subfolder/second_dylib.dylib",
            "First App.app/Contents/Resources/app_packages/subfolder/second_so.so",
            "First App.app/Contents/Resources/app_packages/unknown.binary",
        ]


@pytest.mark.skipif(
    platform.system() != "Darwin",
    reason="ditto tests can only be performed on macOS",
)
def test_ditto_failure(
    package_command,
    first_app_with_binaries,
    tmp_path,
):
    """If ditto fails, an error is raised."""
    app_path = (
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "macos"
        / "app"
        / "First App.app"
    )
    archive_path = tmp_path / "base_path/build/first-app/macos/app/archive.zip"

    # Reconnect the subprocess.run with its live implementation
    package_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        1, cmd=["ditto"]
    )

    # Build the ditto archive
    with pytest.raises(BriefcaseCommandError, match=r"Unable to archive First App.app"):
        package_command.ditto_archive(app_path, archive_path)



import json
import os
import sys
import traceback


def start_remote_debugger():
    try:
        # check verbose output
        verbose = os.environ.get("BRIEFCASE_DEBUG", "0") == "1"

        # reading config
        config_str = os.environ.get("BRIEFCASE_DEBUGGER", None)

        # skip debugger if no config is set
        if config_str is None:
            if verbose:
                print(
                    "No 'BRIEFCASE_DEBUGGER' environment variable found. Debugger not starting."
                )
            return  # If BRIEFCASE_DEBUGGER is not set, this packages does nothing...

        if verbose:
            print(f"'BRIEFCASE_DEBUGGER'={config_str}")

        # Parsing config json
        config = json.loads(config_str)

        # start debugger
        print("Starting remote debugger...")
        if config["debugger"] == "debugpy":
            from briefcase_debugger.debugpy import start_debugpy

            start_debugpy(config, verbose)
        elif config["debugger"] == "pdb":
            from briefcase_debugger.pdb import start_pdb

            start_pdb(config, verbose)
        else:
            raise ValueError(f"Unknown debugger '{config['debugger']}'")
    except Exception:
        # Show exception and stop the whole application when an error occurs
        print(traceback.format_exc())
        sys.exit(-1)



import sys
from collections import namedtuple
from datetime import datetime
from unittest import mock

import pytest

import briefcase


def _create_version_info(major, minor, patch=0):
    Version = namedtuple("Version", "major minor patch")
    return Version(major, minor, patch)


@pytest.mark.parametrize(
    ("minor_version", "today", "is_valid"),
    [
        (8, datetime(2025, 10, 1), False),  # after EOL
        (9, datetime(2025, 10, 1), True),  # on EOL
        (9, datetime(2025, 10, 2), False),  # after EOL
        (10, datetime(2026, 9, 30), True),  # before EOL
        (10, datetime(2026, 10, 2), False),  # after EOL
        (14, datetime(2030, 9, 30), True),  # before EOL
        (14, datetime(2030, 10, 2), False),  # after EOL
    ],
)
def test_valid_python_version(
    minor_version, today, is_valid, base_command, monkeypatch
):
    """A warning is produced if the Python version is past its EOL."""

    version_info = _create_version_info(3, minor_version)
    sys_mock = mock.MagicMock(wraps=sys)
    sys_mock.version_info = version_info
    monkeypatch.setattr(briefcase.commands.base, "sys", sys_mock)

    datetime_mock = mock.MagicMock(wraps=datetime)
    datetime_mock.today.return_value = today
    monkeypatch.setattr(briefcase.commands.base, "datetime", datetime_mock)

    assert base_command.validate_python_version() is is_valid



import shlex
import sys
from unittest import mock

import pytest

from briefcase import __version__, cmdline
from briefcase.commands import ConvertCommand, DevCommand, NewCommand, UpgradeCommand
from briefcase.console import Console, LogLevel
from briefcase.exceptions import (
    InvalidFormatError,
    InvalidPlatformError,
    NoCommandError,
    UnsupportedCommandError,
)
from briefcase.platforms.linux.system import LinuxSystemCreateCommand
from briefcase.platforms.macOS.app import (
    macOSAppCreateCommand,
    macOSAppPublishCommand,
    macOSAppRunCommand,
)
from briefcase.platforms.web.static import StaticWebDevCommand
from briefcase.platforms.windows.app import WindowsAppCreateCommand


@pytest.fixture
def console() -> Console:
    console = Console()
    yield console
    console.close()


@pytest.fixture
def macOS_console(console, monkeypatch, tmp_path) -> Console:
    # Pretend we're on macOS, regardless of where the tests run.
    monkeypatch.setattr(sys, "platform", "darwin")
    # Force the home directory to a temp path. This is needed because faking macOS
    # causes problems with Platformdirs evaluating the user paths.
    monkeypatch.setenv("BRIEFCASE_HOME", str(tmp_path))

    return console


def do_cmdline_parse(args: str, console: Console):
    """Simulate process to parse command line."""
    Command, extra_cmdline = cmdline.parse_cmdline(shlex.split(args))
    cmd = Command(console=console)
    options, overrides = cmd.parse_options(extra=extra_cmdline)
    return cmd, options, overrides


def test_empty():
    """``briefcase`` returns basic usage."""
    with pytest.raises(NoCommandError, match=r"usage: briefcase") as excinfo:
        cmdline.parse_cmdline([])

    assert excinfo.value.msg.startswith(
        "usage: briefcase [-h] <command> [<platform>] [<format>] ...\n"
        "\n"
        "Briefcase is a tool for converting a Python project into a standalone native\n"
        "application for distribution.\n"
        "\n"
        "Commands:\n"
    )


def test_help_only():
    """``briefcase -h`` returns basic usage."""
    with pytest.raises(NoCommandError, match=r"usage: briefcase") as excinfo:
        cmdline.parse_cmdline(["-h"])

    assert excinfo.value.msg.startswith(
        "usage: briefcase [-h] <command> [<platform>] [<format>] ...\n"
        "\n"
        "Briefcase is a tool for converting a Python project into a standalone native\n"
        "application for distribution.\n"
        "\n"
        "Commands:\n"
    )


def test_version_only(capsys):
    """``briefcase -V`` returns current version."""
    with pytest.raises(SystemExit) as excinfo:
        cmdline.parse_cmdline(["-V"])

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Version is displayed.
    output = capsys.readouterr().out
    assert output == f"{__version__}\n"


def test_unknown_command():
    """``briefcase foobar`` fails as an invalid command."""
    with pytest.raises(SystemExit) as excinfo:
        cmdline.parse_cmdline(["foobar"])

    assert excinfo.value.code == 2
    assert excinfo.value.__context__.argument_name == "command"
    assert excinfo.value.__context__.message.startswith(
        "invalid choice: 'foobar' (choose from"
    )


@pytest.mark.parametrize(
    ("cmdline", "expected_options", "expected_overrides"),
    [
        (
            "new",
            {
                "template": None,
                "template_branch": None,
                "project_overrides": None,
            },
            {},
        ),
        (
            (
                "new --template=path/to/template --template-branch=experiment -C"
                " version=\\'1.2.3\\' -C other=42"
            ),
            {
                "template": "path/to/template",
                "template_branch": "experiment",
                "project_overrides": None,
            },
            {
                "version": "1.2.3",
                "other": 42,
            },
        ),
    ],
)
def test_new_command(macOS_console, cmdline, expected_options, expected_overrides):
    """``briefcase new`` returns the New command."""
    cmd, options, overrides = do_cmdline_parse(cmdline, macOS_console)

    assert isinstance(cmd, NewCommand)
    assert cmd.platform == "all"
    assert cmd.output_format == ""
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == expected_options
    assert overrides == expected_overrides


@pytest.mark.parametrize(
    ("cmdline", "expected_options", "expected_overrides"),
    [
        (
            "convert",
            {
                "template": None,
                "template_branch": None,
                "project_overrides": None,
            },
            {},
        ),
        (
            (
                "convert --template=path/to/template --template-branch=experiment -C"
                " version=\\'1.2.3\\' -C other=42"
            ),
            {
                "template": "path/to/template",
                "template_branch": "experiment",
                "project_overrides": None,
            },
            {
                "version": "1.2.3",
                "other": 42,
            },
        ),
    ],
)
def test_convert_command(console, cmdline, expected_options, expected_overrides):
    """``briefcase convert`` returns the Convert command."""
    cmd, options, overrides = do_cmdline_parse(cmdline, console)

    assert isinstance(cmd, ConvertCommand)
    assert cmd.platform == "all"
    assert cmd.output_format == ""
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == expected_options
    assert overrides == expected_overrides


# Common tests for dev and run commands.
def dev_run_parameters(command):
    return [
        (f"{command} {args}", expected, overrides)
        for args, expected, overrides in [
            ("", {}, {}),
            ("-r", {"update_requirements": True}, {}),
            (
                "-r -C version=\\'1.2.3\\' -C other=42",
                {"update_requirements": True},
                {
                    "version": "1.2.3",
                    "other": 42,
                },
            ),
            ("--update-requirements", {"update_requirements": True}, {}),
            ("--test", {"test_mode": True}, {}),
            ("--test -r", {"test_mode": True, "update_requirements": True}, {}),
            ("--", {}, {}),
            ("-- ''", {"passthrough": [""]}, {}),
            ("-- --test", {"passthrough": ["--test"]}, {}),
            ("--test -- --test", {"test_mode": True, "passthrough": ["--test"]}, {}),
            ("--test -- -r", {"test_mode": True, "passthrough": ["-r"]}, {}),
            (
                "-r -- --test",
                {"update_requirements": True, "passthrough": ["--test"]},
                {},
            ),
            ("-- -y --no maybe", {"passthrough": ["-y", "--no", "maybe"]}, {}),
            (
                "--test -- -y --no maybe",
                {"test_mode": True, "passthrough": ["-y", "--no", "maybe"]},
                {},
            ),
        ]
    ]


@pytest.mark.parametrize(
    ("cmdline", "expected_output_format", "expected_options", "expected_overrides"),
    [
        *[(c, "app", o, ov) for c, o, ov in dev_run_parameters("dev")],
        ("dev --no-run", "app", {"run_app": False}, {}),
        ("dev --no-isolation", "app", {"isolated": False}, {}),
        ("dev macOS", "app", {}, {}),
        ("dev macOS Xcode", "Xcode", {}, {}),
    ],
)
def test_dev_command(
    macOS_console,
    cmdline,
    expected_output_format,
    expected_options,
    expected_overrides,
):
    """``briefcase dev`` returns the Dev command."""
    cmd, options, overrides = do_cmdline_parse(cmdline, macOS_console)

    assert isinstance(cmd, DevCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == expected_output_format
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {
        "appname": None,
        "isolated": True,
        "update_requirements": False,
        "run_app": True,
        "test_mode": False,
        "passthrough": [],
        **expected_options,
    }
    assert overrides == expected_overrides


@pytest.mark.parametrize(
    ("cmdline", "expected_output_format", "expected_options", "expected_overrides"),
    [
        ("dev web", "static", {}, {}),
        ("dev web --no-isolation", "static", {"isolated": False}, {}),
    ],
)
def test_web_dev_command(
    macOS_console,
    cmdline,
    expected_output_format,
    expected_options,
    expected_overrides,
):
    """``briefcase dev`` returns the Dev command."""
    cmd, options, overrides = do_cmdline_parse(cmdline, macOS_console)

    assert isinstance(cmd, StaticWebDevCommand)
    assert cmd.platform == "web"
    assert cmd.output_format == expected_output_format
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {
        "appname": None,
        "update_requirements": False,
        "run_app": True,
        "test_mode": False,
        "passthrough": [],
        "isolated": True,
        **expected_options,
    }
    assert overrides == expected_overrides


@pytest.mark.parametrize(
    ("cmdline", "expected_options", "expected_overrides"),
    [
        *dev_run_parameters("run"),
        ("run -u", {"update": True}, {}),
        ("run --update", {"update": True}, {}),
        ("run --update-resources", {"update_resources": True}, {}),
        ("run --update-support", {"update_support": True}, {}),
        ("run --update-stub", {"update_stub": True}, {}),
        ("run --no-update", {"no_update": True}, {}),
    ],
)
def test_run_command(
    macOS_console,
    cmdline,
    expected_options,
    expected_overrides,
):
    """``briefcase run`` returns the Run command for the correct platform."""
    cmd, options, overrides = do_cmdline_parse(cmdline, macOS_console)

    assert isinstance(cmd, macOSAppRunCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {
        "appname": None,
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "debugger": None,
        "debugger_host": "localhost",
        "debugger_port": 5678,
        "passthrough": [],
        **expected_options,
    }
    assert overrides == expected_overrides


@pytest.mark.parametrize(
    ("cmdline", "expected_options", "expected_overrides"),
    [
        (
            "upgrade",
            {
                "list_tools": False,
                "tool_list": [],
            },
            {},
        ),
        (
            "upgrade -C version=\"'1.2.3'\" -C other=42",
            {
                "list_tools": False,
                "tool_list": [],
            },
            {
                "version": "1.2.3",
                "other": 42,
            },
        ),
    ],
)
def test_upgrade_command(
    macOS_console,
    cmdline,
    expected_options,
    expected_overrides,
):
    """``briefcase upgrade`` returns the upgrade command."""
    cmd, options, overrides = do_cmdline_parse(cmdline, macOS_console)

    assert isinstance(cmd, UpgradeCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == ""
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == expected_options
    assert overrides == expected_overrides


def test_bare_command(macOS_console):
    """``briefcase create`` returns the macOS create app command."""
    cmd, options, overrides = do_cmdline_parse("create", macOS_console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}
    assert overrides == {}


@pytest.mark.skipif(sys.platform != "linux", reason="requires Linux")
def test_linux_default(console):
    """``briefcase create`` returns the linux create system command on Linux."""

    cmd, options, _overrides = do_cmdline_parse("create", console)

    assert isinstance(cmd, LinuxSystemCreateCommand)
    assert cmd.platform == "linux"
    assert cmd.output_format == "system"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}


@pytest.mark.skipif(sys.platform != "darwin", reason="requires macOS")
def test_macOS_default(console):
    """``briefcase create`` returns the macOS create command on Linux."""

    cmd, options, overrides = do_cmdline_parse("create", console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}
    assert overrides == {}


@pytest.mark.skipif(sys.platform != "win32", reason="requires Windows")
def test_windows_default(console):
    """``briefcase create`` returns the Windows create app command on Windows."""

    cmd, options, overrides = do_cmdline_parse("create", console)

    assert isinstance(cmd, WindowsAppCreateCommand)
    assert cmd.platform == "windows"
    assert cmd.output_format == "app"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}
    assert overrides == {}


def test_bare_command_help(capsys, macOS_console):
    """``briefcase create -h`` returns the macOS create app command help."""
    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create -h", macOS_console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Help message is for default platform and format
    output = capsys.readouterr().out
    assert output.startswith(
        "usage: briefcase create macOS app [-h] [-C KEY=VALUE] [-v] [-V] [--no-input]\n"
        "                                  [--log] [-a APP_NAME]\n"
        "\n"
        "Create and populate a macOS app.\n"
    )


def test_bare_command_version(capsys, console):
    """``briefcase create -V`` returns the version."""
    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create -V", console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Version is displayed.
    output = capsys.readouterr().out
    assert output == f"{__version__}\n"


def test_command_unknown_platform(macOS_console):
    """``briefcase create foobar`` raises an unknown platform error."""
    expected_exc_regex = r"Invalid platform 'foobar'; \(choose from: .*\)"
    with pytest.raises(InvalidPlatformError, match=expected_exc_regex):
        do_cmdline_parse("create foobar", macOS_console)


def test_command_explicit_platform(macOS_console):
    """``briefcase create linux`` returns linux create app command."""
    cmd, options, overrides = do_cmdline_parse("create linux", macOS_console)

    assert isinstance(cmd, LinuxSystemCreateCommand)
    assert cmd.platform == "linux"
    assert cmd.output_format == "system"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}
    assert overrides == {}


def test_command_explicit_platform_case_handling(macOS_console):
    """``briefcase create macOS`` returns macOS create app command."""
    # This is all lower case; the command normalizes to macOS
    cmd, options, overrides = do_cmdline_parse("create macOS", macOS_console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}
    assert overrides == {}


def test_command_explicit_platform_help(capsys, macOS_console):
    """``briefcase create macOS -h`` returns the macOS create app command help."""
    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create macOS -h", macOS_console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Help message is for default platform and format
    output = capsys.readouterr().out
    assert output.startswith(
        "usage: briefcase create macOS app [-h] [-C KEY=VALUE] [-v] [-V] [--no-input]\n"
        "                                  [--log] [-a APP_NAME]\n"
        "\n"
        "Create and populate a macOS app.\n"
    )


def test_command_explicit_format(macOS_console):
    """``briefcase create macOS app`` returns the macOS create app command."""
    cmd, options, overrides = do_cmdline_parse("create macOS app", macOS_console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}
    assert overrides == {}


def test_command_unknown_format(macOS_console):
    """``briefcase create macOS foobar`` returns an invalid format error."""
    expected_exc_regex = r"Invalid format 'foobar'; \(choose from: app, Xcode\)"
    with pytest.raises(InvalidFormatError, match=expected_exc_regex):
        do_cmdline_parse("create macOS foobar", macOS_console)


def test_command_explicit_unsupported_format(monkeypatch, macOS_console):
    """``briefcase create macOS homebrew`` raises an error because the format isn't
    supported (yet)"""
    # Mock the output formats to include a "homebrew" backend with no commands.
    monkeypatch.setattr(
        cmdline,
        "get_output_formats",
        mock.MagicMock(return_value={"homebrew": None}),
    )

    with pytest.raises(
        UnsupportedCommandError,
        match=(
            r"The create command for the macOS homebrew format has not been implemented"
            r" \(yet!\)."
        ),
    ):
        do_cmdline_parse("create macOS homebrew", macOS_console)


def test_command_explicit_format_help(capsys, macOS_console):
    """``briefcase create macOS app -h`` returns the macOS create app help."""
    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("create macOS app -h", macOS_console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 0
    # Help message is for default platform, but app format
    output = capsys.readouterr().out
    assert output.startswith(
        "usage: briefcase create macOS app [-h] [-C KEY=VALUE] [-v] [-V] [--no-input]\n"
        "                                  [--log] [-a APP_NAME]\n"
        "\n"
        "Create and populate a macOS app.\n"
    )


def test_command_disable_input(macOS_console):
    """``briefcase create --no-input`` disables console input."""
    cmd, options, overrides = do_cmdline_parse("create --no-input", macOS_console)

    assert isinstance(cmd, macOSAppCreateCommand)
    assert cmd.platform == "macOS"
    assert cmd.output_format == "app"
    assert not cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {}
    assert overrides == {}


def test_command_options(capsys, macOS_console):
    """Commands can provide their own arguments."""
    # Invoke a command that is known to have its own custom arguments
    # (In this case, the channel argument for publication)
    cmd, options, overrides = do_cmdline_parse("publish macos app -c s3", macOS_console)

    assert isinstance(cmd, macOSAppPublishCommand)
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {"update": False, "packaging_format": "app", "channel": "s3"}
    assert overrides == {}


def test_command_overrides(capsys, macOS_console):
    """Configuration overrides can be specified."""
    # Invoke a command that is known to have its own custom arguments
    # (In this case, the channel argument for publication)
    cmd, options, overrides = do_cmdline_parse(
        "publish macos app -C version=\"'1.2.3'\" -C extra=42", macOS_console
    )

    assert isinstance(cmd, macOSAppPublishCommand)
    assert cmd.console.input_enabled
    assert cmd.console.verbosity == LogLevel.INFO
    assert options == {"update": False, "packaging_format": "app", "channel": None}
    assert overrides == {
        "version": "1.2.3",
        "extra": 42,
    }


def test_unknown_command_options(capsys, macOS_console):
    """Commands can provide their own arguments."""
    # Invoke a command but provide an option. that isn't defined
    with pytest.raises(SystemExit) as excinfo:
        do_cmdline_parse("publish macOS xcode -x foobar", macOS_console)

    # Normal exit due to displaying help
    assert excinfo.value.code == 2
    # Help message is for default platform and format
    output = capsys.readouterr().err

    assert output.startswith(
        "usage: briefcase publish macOS Xcode [-h] [-C KEY=VALUE] [-v] [-V]\n"
        "                                     [--no-input] [--log] [-a APP_NAME] [-u]\n"
        "                                     [-p {Xcode}] [-c CHANNEL]\n"
        "briefcase publish macOS Xcode: error: unrecognized arguments: -x foobar"
    )


@pytest.mark.parametrize(
    "cmdline",
    [
        "dev iOS",
        "dev Android",
    ],
)
def test_dev_command_unsupported_platform(cmdline, console):
    """Unsupported platforms should raise an UnsupportedCommandError."""

    with pytest.raises(UnsupportedCommandError) as e:
        do_cmdline_parse(cmdline, console)

    msg = str(e.value)
    assert "dev command for the" in msg
    assert "not been implemented" in msg



import pytest
from packaging.version import Version

from briefcase.config import GlobalConfig
from briefcase.exceptions import BriefcaseConfigError

# Examples of valid versions; input, parsed
VALID_VERSIONS = [
    ("1", "1"),  # Single digit
    ("1.0", "1.0"),  # Two digits
    ("1.2.3", "1.2.3"),  # Semver digits
    ("1.2.3.4", "1.2.3.4"),  # Keep going...
    ("v1.2.3", "1.2.3"),  # v prefix
    ("V1.2.3", "1.2.3"),  # v prefix, capitalized
    ("2025.9.5", "2025.9.5"),  # Calver
    ("2025.09.05", "2025.9.5"),  # Calver with leading zeros
    ("1.2.3.dev42", "1.2.3.dev42"),  # dev
    ("1.2.3a42", "1.2.3a42"),  # alpha
    ("1.2.3b42", "1.2.3b42"),  # beta
    ("1.2.3rc42", "1.2.3rc42"),  # release candidate
    ("1.2.3.post42", "1.2.3.post42"),  # post release
    ("1.2.3.a42", "1.2.3a42"),  # alpha with period
    ("1.2.3.b42", "1.2.3b42"),  # beta with period
    ("1.2.3.rc42", "1.2.3rc42"),  # release candidate with period
    ("1.2.3post42", "1.2.3.post42"),  # post release without period
    ("1.2.3.post37.dev42", "1.2.3.post37.dev42"),  # dev post release
    ("1.2.3.DEV42", "1.2.3.dev42"),  # dev, capitalized
    ("1.2.3A42", "1.2.3a42"),  # alpha, capitalized
    ("1.2.3B42", "1.2.3b42"),  # beta, capitalized
    ("1.2.3RC42", "1.2.3rc42"),  # release candidate, capitalized
    ("1.2.3.POST42", "1.2.3.post42"),  # post release, capitalized
    ("1.2.3alpha42", "1.2.3a42"),  # alpha, full spelling
    ("1.2.3beta42", "1.2.3b42"),  # beta, full spelling
    ("1.2.3c42", "1.2.3rc42"),  # release candidate, alternate spelling
    ("1.2.3pre42", "1.2.3rc42"),  # release candidate, alternate spelling
    ("1.2.3preview42", "1.2.3rc42"),  # release candidate, alternate spelling
    ("1.2.3-42", "1.2.3.post42"),  # post release, alternate spelling
    ("1.2.3-r42", "1.2.3.post42"),  # post release, alternate spelling
    ("1.2.3a", "1.2.3a0"),  # implied alpha version
    ("1.2.3b", "1.2.3b0"),  # implied beta version
    ("1.2.3rc", "1.2.3rc0"),  # implied release candidate version
    ("1.2.3.post", "1.2.3.post0"),  # implied post-release version
    ("1.2.3a05", "1.2.3a5"),  # leading zero trimming in postversion digits
    ("1.2.3-dev5", "1.2.3.dev5"),  # hyphen in postversion digits
    ("1.2.3_dev5", "1.2.3.dev5"),  # underscore in postversion digits
    ("1.2.3a-5", "1.2.3a5"),  # hyphen and letter in postversion digits
    ("1.2.3a_5", "1.2.3a5"),  # underscore and letter in postversion digits
    ("2!1.2.3", "2!1.2.3"),  # Epoch
    ("1.2.3+ubuntu.1", "1.2.3+ubuntu.1"),  # Local version segment
    (
        "v2!1.2.3rc37.post42+ubuntu.1",
        "2!1.2.3rc37.post42+ubuntu.1",
    ),  # all the options
    (Version("1.2.3"), "1.2.3"),  # Version object as input
]


def test_minimal_GlobalConfig():
    """A simple config can be defined."""
    config = GlobalConfig(
        project_name="My Project",
        version="1.2.3",
        bundle="org.beeware",
        license={"file": "LICENSE"},
    )

    # The basic properties have been set.
    assert config.project_name == "My Project"
    assert config.version == Version("1.2.3")
    assert config.bundle == "org.beeware"

    assert repr(config) == "<My Project v1.2.3 GlobalConfig>"


def test_extra_attrs():
    """A config can contain attributes in addition to those required."""
    config = GlobalConfig(
        project_name="My Project",
        version="1.2.3",
        bundle="org.beeware",
        url="https://example.com",
        author="Jane Smith",
        author_email="jane@example.com",
        first="value 1",
        second=42,
        license={"file": "LICENSE"},
    )

    # The basic properties have been set.
    assert config.project_name == "My Project"
    assert config.version == Version("1.2.3")
    assert config.bundle == "org.beeware"
    assert config.url == "https://example.com"
    assert config.author == "Jane Smith"
    assert config.author_email == "jane@example.com"

    # Explicit additional properties have been set
    assert config.first == "value 1"
    assert config.second == 42

    # An attribute that wasn't provided raises an error
    with pytest.raises(AttributeError):
        _ = config.unknown


@pytest.mark.parametrize(("input", "expected"), VALID_VERSIONS)
def test_valid_app_version(input, expected):
    config = GlobalConfig(
        project_name="My Project",
        version=input,
        bundle="org.beeware",
        license={"file": "LICENSE"},
    )

    # Version is parsed as an equivalent Version object
    assert config.version == Version(expected)
    # Serialized Version is the expected value
    assert str(config.version) == expected


def test_invalid_app_version():
    with pytest.raises(
        BriefcaseConfigError, match=r"Version number \(foobar\) is not valid\."
    ):
        GlobalConfig(
            project_name="My Project",
            version="foobar",
            bundle="org.beeware",
            license={"file": "LICENSE"},
        )



from unittest import mock

import pytest

from briefcase.platforms.macOS import SigningIdentity
from briefcase.platforms.macOS.app import macOSAppBuildCommand


@pytest.fixture
def build_command(dummy_console, tmp_path):
    command = macOSAppBuildCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )

    command.select_identity = mock.MagicMock()
    command.sign_app = mock.MagicMock()
    command.sign_file = mock.MagicMock()
    command.verify_not_on_icloud = mock.MagicMock()

    return command


@pytest.mark.parametrize(
    "extra_args", [{}, {"adhoc_sign": True}, {"adhoc_sign": False}]
)
@pytest.mark.parametrize("universal_build", [True, False])
@pytest.mark.parametrize("pre_existing", [True, False])
@pytest.mark.parametrize("console_app", [True, False])
def test_build_app(
    build_command,
    first_app_with_binaries,
    console_app,
    universal_build,
    pre_existing,
    extra_args,
    tmp_path,
):
    """A macOS App is ad-hoc signed as part of the build process."""
    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    first_app_with_binaries.universal_build = universal_build
    first_app_with_binaries.console_app = console_app
    build_command.tools.host_arch = "gothic"

    exec_path = bundle_path / "First App.app/Contents/MacOS"
    if not pre_existing:
        # If this not a pre-existing app, the stub has the original name
        (exec_path / "First App").rename(exec_path / "Stub")

    # Mock the thin command so we can confirm if it was invoked.
    build_command.ensure_thin_binary = mock.Mock()

    # Build the app
    base_args = {"test_mode": False}
    kwargs = base_args | extra_args
    build_command.build_app(first_app_with_binaries, **kwargs)

    # We verified we aren't on iCloud
    build_command.verify_not_on_icloud.assert_called_once_with(first_app_with_binaries)

    # The stub binary has been renamed
    assert not (exec_path / "Stub").is_file()
    assert (exec_path / "First App").is_file()

    # Only thin if this is a non-universal app
    if universal_build:
        build_command.ensure_thin_binary.assert_not_called()
    else:
        build_command.ensure_thin_binary.assert_called_once_with(
            exec_path / "First App",
            arch="gothic",
        )

    # Verify that a request has been made to sign the app, but only when it is as part of a direct build command
    if "adhoc_sign" in kwargs:
        # build command was called as part of package command.
        # expect no signing now since it will happen during packaging
        build_command.sign_app.assert_not_called()
    else:
        # build command was called directly so expect signing
        build_command.sign_app.assert_called_once_with(
            app=first_app_with_binaries,
            identity=SigningIdentity(),
        )

    # No request to select a signing identity was made
    build_command.select_identity.assert_not_called()

    # No attempt was made to sign a specific file;
    # This ignores the calls that would have been made transitively
    # by calling sign_app()
    build_command.sign_file.assert_not_called()



from collections.abc import Collection
from unittest.mock import MagicMock

import pytest

import briefcase.commands.new
from briefcase.bootstraps import (
    ConsoleBootstrap,
    EmptyBootstrap,
    PygameGuiBootstrap,
    PySide6GuiBootstrap,
    TogaGuiBootstrap,
)


@pytest.fixture
def mock_builtin_bootstraps():
    return {
        "Toga": TogaGuiBootstrap,
        "Console": ConsoleBootstrap,
        "PySide6": PySide6GuiBootstrap,
        "Pygame": PygameGuiBootstrap,
    }


def test_toga_bootstrap(new_command):
    """Context can be requested from the Toga bootstrap."""

    context = new_command.build_gui_context(
        TogaGuiBootstrap(
            new_command.console,
            {
                "app_name": "myapplication",
                "author": "Grace Hopper",
            },
        ),
        project_overrides={},
    )

    assert context == {
        "app_source": '''\
import toga
from toga.style.pack import COLUMN, ROW


class {{ cookiecutter.class_name }}(toga.App):
    def startup(self):
        """Construct and show the Toga application.

        Usually, you would add your application to a main content box.
        We then create a main window (with a name matching the app), and
        show the main window.
        """
        main_box = toga.Box()

        self.main_window = toga.MainWindow(title=self.formal_name)
        self.main_window.content = main_box
        self.main_window.show()


def main():
    return {{ cookiecutter.class_name }}()
''',
        "app_start_source": """\
from {{ cookiecutter.module_name }}.app import main

if __name__ == "__main__":
    main().main_loop()
""",
        "pyproject_table_briefcase_app_extra_content": """
requires = [
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        "pyproject_table_macOS": """\
universal_build = true
requires = [
    "toga-cocoa~=0.5.0",
    "std-nslog~=1.0.3",
]
""",
        "pyproject_table_linux": """\
requires = [
    "toga-gtk~=0.5.0",
    # PyGObject 3.52.1 enforces a requirement on libgirepository-2.0-dev. This library
    # isn't available on Debian 12/Ubuntu 22.04. If you don't need to support those (or
    # older) releases, you can remove this version pin. See beeware/toga#3143.
    "pygobject < 3.52.1",
]
""",
        "pyproject_table_linux_system_debian": """\
system_requires = [
    # Needed to compile pycairo wheel
    "libcairo2-dev",
    # One of the following two packages are needed to compile PyGObject wheel. If you
    # remove the pygobject pin in the requires list, you should also change to the
    # version 2.0 of the girepository library. See beeware/toga#3143.
    "libgirepository1.0-dev",
    # "libgirepository-2.0-dev",
]

system_runtime_requires = [
    # Needed to provide GTK and its GI bindings
    "gir1.2-gtk-3.0",
    # One of the following two packages are needed to use PyGObject at runtime. If you
    # remove the pygobject pin in the requires list, you should also change to the
    # version 2.0 of the girepository library. See beeware/toga#3143.
    "libgirepository-1.0-1",
    # "libgirepository-2.0-0",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3-module",
    # Needed to provide WebKit2 at runtime
    # Note: Debian 11 requires gir1.2-webkit2-4.0 instead
    # "gir1.2-webkit2-4.1",
]
""",
        "pyproject_table_linux_system_rhel": """\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo-gobject-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
]

system_runtime_requires = [
    # Needed to support Python bindings to GTK
    "gobject-introspection",
    # Needed to provide GTK
    "gtk3",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3",
    # Needed to provide WebKit2 at runtime
    # "webkit2gtk3",
]
""",
        "pyproject_table_linux_system_suse": """\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
]

system_runtime_requires = [
    # Needed to provide GTK
    "gtk3",
    # Needed to support Python bindings to GTK
    "gobject-introspection", "typelib(Gtk) = 3.0",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3-module",
    # Needed to provide WebKit2 at runtime
    # "libwebkit2gtk3", "typelib(WebKit2)",
]
""",
        "pyproject_table_linux_system_arch": """\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo",
    # Needed to compile PyGObject wheel
    "gobject-introspection",
    # Runtime dependencies that need to exist so that the
    # Arch package passes final validation.
    # Needed to provide GTK
    "gtk3",
    # Dependencies that GTK looks for at runtime
    "libcanberra",
    # Needed to provide WebKit2
    # "webkit2gtk",
]

system_runtime_requires = [
    # Needed to provide GTK
    "gtk3",
    # Needed to provide PyGObject bindings
    "gobject-introspection-runtime",
    # Dependencies that GTK looks for at runtime
    "libcanberra",
    # Needed to provide WebKit2 at runtime
    # "webkit2gtk",
]
""",
        "pyproject_table_linux_appimage": """\
manylinux = "manylinux_2_28"

system_requires = [
    # Needed to compile pycairo wheel
    "cairo-gobject-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
    # Needed to provide GTK
    "gtk3-devel",
    # Dependencies that GTK looks for at runtime, that need to be
    # in the build environment to be picked up by linuxdeploy
    "libcanberra-gtk3",
    "PackageKit-gtk3-module",
    "gvfs-client",
]

linuxdeploy_plugins = [
    "DEPLOY_GTK_VERSION=3 gtk",
]
""",
        "pyproject_table_linux_flatpak": """\
flatpak_runtime = "org.gnome.Platform"
flatpak_runtime_version = "48"
flatpak_sdk = "org.gnome.Sdk"
""",
        "pyproject_table_windows": """\
requires = [
    "toga-winforms~=0.5.0",
]
""",
        "pyproject_table_iOS": """\
requires = [
    "toga-iOS~=0.5.0",
    "std-nslog~=1.0.3",
]
""",
        "pyproject_table_android": '''\
requires = [
    "toga-android~=0.5.0",
]

base_theme = "Theme.MaterialComponents.Light.DarkActionBar"

build_gradle_dependencies = [
    "com.google.android.material:material:1.13.0",
    # Needed for DetailedList
    # "androidx.swiperefreshlayout:swiperefreshlayout:1.1.0",
    # Needed for MapView
    # "org.osmdroid:osmdroid-android:6.1.20",
    # Needed for WebView
    # "androidx.webkit:webkit:1.15.0",
]

# Needed for WebView
# build_gradle_extra_content="""
# chaquopy.defaultConfig.staticProxy("toga_android.widgets.internal.webview")
# """
''',
        "pyproject_table_web": """\
requires = [
    "toga-web~=0.5.0",
]
""",
    }


def test_console_bootstrap(new_command):
    """Context can be requested from the Console bootstrap."""

    context = new_command.build_gui_context(
        ConsoleBootstrap(
            new_command.console,
            {
                "app_name": "myapplication",
                "author": "Grace Hopper",
            },
        ),
        project_overrides={},
    )

    assert context == {
        "console_app": True,
        "app_source": (
            """\

def main():
    # Your app logic goes here
    print("Hello, World.")
"""
        ),
        "app_start_source": """\
from {{ cookiecutter.module_name }}.app import main

if __name__ == "__main__":
    main()
""",
        "pyproject_table_briefcase_app_extra_content": """
requires = [
    # Add your cross-platform app requirements here
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        "pyproject_table_macOS": """\
universal_build = true
requires = [
    # Add your macOS-specific app requirements here
]
""",
        "pyproject_table_linux": """\
requires = [
    # Add your Linux-specific app requirements here
]
""",
        "pyproject_table_linux_system_debian": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_system_rhel": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_system_suse": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_system_arch": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_flatpak": """\
flatpak_runtime = "org.freedesktop.Platform"
flatpak_runtime_version = "25.08"
flatpak_sdk = "org.freedesktop.Sdk"
""",
        "pyproject_table_windows": """\
requires = [
    # Add your Windows-specific app requirements here
]
""",
        "pyproject_table_iOS": """\
supported = false
""",
        "pyproject_table_android": """\
supported = false
""",
        "pyproject_table_web": """\
supported = false
""",
    }


def test_pyside6_bootstrap(new_command):
    """Context can be requested from the PySide6 bootstrap."""

    context = new_command.build_gui_context(
        PySide6GuiBootstrap(
            new_command.console,
            {
                "app_name": "myapplication",
                "author": "Grace Hopper",
            },
        ),
        project_overrides={},
    )

    assert context == {
        "app_source": """\
import importlib.metadata
import sys

from PySide6 import QtWidgets


class {{ cookiecutter.class_name }}(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("{{ cookiecutter.app_name }}")
        self.show()


def main():
    # Linux desktop environments use an app's .desktop file to integrate the app
    # in to their application menus. The .desktop file of this app will include
    # the StartupWMClass key, set to app's formal name. This helps associate the
    # app's windows to its menu item.
    #
    # For association to work, any windows of the app must have WMCLASS property
    # set to match the value set in app's desktop file. For PySide6, this is set
    # with setApplicationName().

    # Find the name of the module that was used to start the app
    app_module = sys.modules["__main__"].__package__
    # Retrieve the app's metadata
    metadata = importlib.metadata.metadata(app_module)

    QtWidgets.QApplication.setApplicationName(metadata["Formal-Name"])

    app = QtWidgets.QApplication(sys.argv)
    main_window = {{ cookiecutter.class_name }}()
    sys.exit(app.exec())
""",
        "pyproject_table_briefcase_app_extra_content": """
requires = [
    "PySide6-Essentials~=6.8",
    # "PySide6-Addons~=6.8",
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        "pyproject_table_macOS": """\
universal_build = true
# As of Pyside 6.8, PySide enforces a macOS 12 minimum on wheels.
min_os_version = "12.0"
requires = [
    "std-nslog~=1.0.3",
]
""",
        "pyproject_table_linux": """\
requires = [
]
""",
        "pyproject_table_linux_system_debian": """\
system_requires = [
]

system_runtime_requires = [
    "libgl1",
    "libqt6dbus6",
    "libqt6gui6",
    "libxcb-cursor0",
    # The following packages are explicit dependencies on recent Debian-based
    # distros, but not on Ubuntu 22.04.
    "libxcb-icccm4",
    "libxcb-keysyms1",
    "libxcb-shape0",
    "libxkbcommon-x11-0",
]
""",
        "pyproject_table_linux_system_rhel": """\
system_requires = [
]

system_runtime_requires = [
    "qt6-qtbase-gui",
]
""",
        "pyproject_table_linux_system_suse": """\
system_requires = [
]

system_runtime_requires = [
    "libgthread-2_0-0",
    "libQt6Gui6",
]
""",
        "pyproject_table_linux_system_arch": """\
system_requires = [
]

system_runtime_requires = [
    "qt6-base",
]
""",
        "pyproject_table_linux_appimage": """\
manylinux = "manylinux_2_28"

system_requires = [
# ?? FIXME
]

linuxdeploy_plugins = [
]
""",
        "pyproject_table_linux_flatpak": """\
flatpak_runtime = "org.kde.Platform"
flatpak_runtime_version = "6.9"
flatpak_sdk = "org.kde.Sdk"
""",
        "pyproject_table_windows": """\
requires = [
]
""",
        "pyproject_table_iOS": """\
supported = false
""",
        "pyproject_table_android": """\
supported = false
""",
        "pyproject_table_web": """\
supported = false
""",
    }


def test_pygame_bootstrap(new_command):
    """Context can be requested from the Pygame bootstrap."""

    context = new_command.build_gui_context(
        PygameGuiBootstrap(
            new_command.console,
            {
                "app_name": "myapplication",
                "author": "Grace Hopper",
            },
        ),
        project_overrides={},
    )

    assert context == {
        "app_source": """\
import importlib.metadata
import os
import sys
from pathlib import Path

import pygame

SCREEN_WIDTH, SCREEN_HEIGHT = 800, 600
WHITE = (255, 255, 255)


def main():
    # Linux desktop environments use an app's .desktop file to integrate the app
    # in to their application menus. The .desktop file of this app will include
    # the StartupWMClass key, set to app's formal name. This helps associate the
    # app's windows to its menu item.
    #
    # For association to work, any windows of the app must have WMCLASS property
    # set to match the value set in app's desktop file. For PyGame, this is set
    # using the SDL_VIDEO_X11_WMCLASS environment variable.

    # Find the name of the module that was used to start the app
    app_module = sys.modules["__main__"].__package__
    # Retrieve the app's metadata
    metadata = importlib.metadata.metadata(app_module)

    os.environ["SDL_VIDEO_X11_WMCLASS"] = metadata["Formal-Name"]

    pygame.init()
    pygame.display.set_caption(metadata["Formal-Name"])
    screen = pygame.display.set_mode((SCREEN_WIDTH, SCREEN_HEIGHT))

    running = True
    while running:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
                break

        screen.fill(WHITE)
        pygame.display.flip()

    pygame.quit()
""",
        "pyproject_table_briefcase_app_extra_content": """
requires = [
    "pygame~=2.6",
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        "pyproject_table_macOS": """\
universal_build = true
requires = [
    "std-nslog~=1.0.3",
]
""",
        "pyproject_table_linux": """\
requires = [
]
""",
        "pyproject_table_linux_system_debian": """\
system_requires = [
]

system_runtime_requires = [
]
""",
        "pyproject_table_linux_system_rhel": """\
system_requires = [
]

system_runtime_requires = [
]
""",
        "pyproject_table_linux_system_suse": """\
system_requires = [
]

system_runtime_requires = [
]
""",
        "pyproject_table_linux_system_arch": """\
system_requires = [
]

system_runtime_requires = [
]
""",
        "pyproject_table_linux_appimage": """\
manylinux = "manylinux_2_28"

system_requires = [
]

linuxdeploy_plugins = [
]
""",
        "pyproject_table_linux_flatpak": """\
flatpak_runtime = "org.freedesktop.Platform"
flatpak_runtime_version = "25.08"
flatpak_sdk = "org.freedesktop.Sdk"
""",
        "pyproject_table_windows": """\
requires = [
]
""",
        "pyproject_table_iOS": """\
supported = false
""",
        "pyproject_table_android": """\
supported = false
""",
        "pyproject_table_web": """\
supported = false
""",
    }


def test_no_bootstrap(new_command):
    """The empty bootstrap is used if no bootstrap is selected."""

    context = new_command.build_gui_context(
        EmptyBootstrap(
            new_command.console,
            {
                "app_name": "myapplication",
                "author": "Grace Hopper",
            },
        ),
        project_overrides={},
    )

    assert context == {
        "app_source": """\

def main():
    # Your app logic goes here
    print("Hello, World.")
""",
        "app_start_source": """\
from {{ cookiecutter.module_name }}.app import main

if __name__ == "__main__":
    main()
""",
        "pyproject_table_briefcase_app_extra_content": """
requires = [
    # Add your cross-platform app requirements here
]
test_requires = [
    # Add your cross-platform test requirements here
]
""",
        "pyproject_table_macOS": """\
universal_build = true
requires = [
    # Add your macOS-specific app requirements here
]
""",
        "pyproject_table_linux": """\
requires = [
    # Add your Linux-specific app requirements here
]
""",
        "pyproject_table_linux_system_debian": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_system_rhel": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_system_suse": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_system_arch": """\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        "pyproject_table_linux_flatpak": """\
flatpak_runtime = "org.freedesktop.Platform"
flatpak_runtime_version = "25.08"
flatpak_sdk = "org.freedesktop.Sdk"
""",
        "pyproject_table_windows": """\
requires = [
    # Add your Windows-specific app requirements here
]
""",
        "pyproject_table_iOS": """\
requires = [
    # Add your iOS-specific app requirements here
]
""",
        "pyproject_table_android": """\
requires = [
    # Add your Android-specific app requirements here
]
""",
        "pyproject_table_web": """\
requires = [
    # Add your web-specific app requirements here
]
""",
    }


def test_custom_bootstrap(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """A context is create for a custom bootstrap."""

    class GuiBootstrap:
        fields: Collection[str] = ["requires", "platform"]

        def __init__(self, console, context):
            self.console = console
            self.context = context

        def extra_context(self, project_overrides):
            return {
                "custom_context": "value",
                "custom_override": project_overrides.pop("custom_override", None),
            }

        def requires(self):
            return "toga"

        def platform(self):
            return "bsd"

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value={
                **mock_builtin_bootstraps,
                "Custom GUI": GuiBootstrap,
            },
        ),
    )

    context = new_command.build_gui_context(
        GuiBootstrap(
            new_command.console,
            {
                "app_name": "myapplication",
                "author": "Grace Hopper",
            },
        ),
        project_overrides={"custom_override": "other"},
    )

    assert context == {
        "custom_context": "value",
        "custom_override": "other",
        "requires": "toga",
        "platform": "bsd",
    }



import os
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.integrations.docker import Docker, DockerAppContext
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.appimage import LinuxAppImageOpenCommand

from ...utils import create_file


@pytest.fixture
def open_command(dummy_console, tmp_path):
    command = LinuxAppImageOpenCommand(
        console=dummy_console,
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.os = MagicMock(spec_set=os)

    # Mock x86_64 for linuxdeploy verification
    command.tools.host_arch = "x86_64"

    # Store the underlying subprocess instance
    command._subprocess = MagicMock(spec_set=Subprocess)
    command.tools.subprocess = command._subprocess

    # Mock existence of linuxdeploy in tools
    create_file(
        command.tools.base_path / "linuxdeploy-x86_64.AppImage",
        content="",
    )

    return command


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows paths aren't converted in Docker context",
)
def test_open_docker(open_command, first_app_config, tmp_path, monkeypatch):
    """Open starts a docker session by default."""

    # Enable docker
    open_command.use_docker = True
    open_command.extra_docker_build_args = []

    # Provide Docker
    monkeypatch.setattr(
        Docker, "_is_user_mapping_enabled", MagicMock(return_value=True)
    )
    open_command.tools.docker = Docker(tools=open_command.tools)
    # Provide Docker app context
    open_command.tools[first_app_config].app_context = DockerAppContext(
        tools=open_command.tools,
        app=first_app_config,
    )
    open_command.tools[first_app_config].app_context.prepare(
        image_tag=f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/linux/appimage/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/linux/appimage",
        host_data_path=tmp_path / "briefcase",
        python_version=f"3.{sys.version_info.minor}",
    )
    # Clear out any calls recorded during preparation
    open_command._subprocess.run.reset_mock()

    # Create the desktop file that would be in the project folder.
    create_file(
        open_command.project_path(first_app_config)
        / "First App.AppDir/com.example.firstapp.desktop",
        "FreeDesktop file",
    )

    # Open the app
    open_command.open_app(first_app_config)

    # The docker session was started
    open_command._subprocess.run.assert_called_once_with(
        args=[
            "docker",
            "run",
            "--rm",
            "-it",
            "--volume",
            f"{open_command.base_path}/build/first-app/linux/appimage:/app:z",
            "--volume",
            f"{open_command.data_path}:/briefcase:z",
            f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
        ],
        stream_output=False,
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.skipif(sys.platform != "linux", reason="Linux specific test")
def test_open_no_docker_linux(open_command, first_app_config, tmp_path):
    """On Linux, Open runs `xdg-open` on the project folder if we specify --no-
    docker."""
    # Create the desktop file that would be in the project folder.
    create_file(
        open_command.project_path(first_app_config)
        / "First App.AppDir/com.example.firstapp.desktop",
        "FreeDesktop file",
    )

    # Disable docker
    open_command.use_docker = False

    open_command(first_app_config)

    open_command.tools.subprocess.Popen.assert_called_once_with(
        [
            "xdg-open",
            tmp_path / "base_path/build/first-app/linux/appimage",
        ]
    )


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
def test_open_no_docker_macOS(open_command, first_app_config, tmp_path):
    """On macOS, Open runs `open` on the project folder if we specify --no-docker."""
    # Create the desktop file that would be in the project folder.
    create_file(
        open_command.project_path(first_app_config)
        / "First App.AppDir/com.example.firstapp.desktop",
        "FreeDesktop file",
    )

    # Disable docker
    open_command.use_docker = False

    open_command(first_app_config)

    open_command.tools.subprocess.Popen.assert_called_once_with(
        [
            "open",
            tmp_path / "base_path/build/first-app/linux/appimage",
        ]
    )



import os
import platform
import shutil
import ssl
import stat
from collections.abc import Iterable, Iterator
from pathlib import Path
from unittest import mock

import httpcore
import httpx
import pytest

from briefcase.exceptions import (
    BadNetworkResourceError,
    MissingNetworkResourceError,
    NetworkFailure,
)
from briefcase.integrations.base import ToolCache

TEMPORARY_DOWNLOAD_FILE_SUFFIX = ".download"


@pytest.fixture
def mock_tools(mock_tools) -> ToolCache:
    mock_tools.httpx = mock.MagicMock(spec_set=httpx)
    # Restore move so the temporary file can be moved after downloaded
    mock_tools.shutil.move = mock.MagicMock(wraps=shutil.move)
    return mock_tools


class _IteratorByteSteam(httpx.SyncByteStream):
    """Shim that satisfies ``httpx.Response`` ``stream`` parameter type.

    Cannot be replaced by any ``Iterable[bytes]`` because the base class requires
    an explicit finalization method ``close``.
    """

    def __init__(self, iterable: Iterable[bytes]) -> None:
        self.iterable = iterable

    def __iter__(self) -> Iterator[bytes]:
        return iter(self.iterable)


def _make_httpx_response(
    *,
    url: str,
    status_code: int,
    stream: list[bytes],
    method: str = "GET",
    headers: dict | None = None,
) -> httpx.Response:
    """Create a real ``httpx.Response`` with key methods wrapped by ``mock.Mock`` for
    spying.

    Wrapped methods:
        response.read
        response.iter_bytes
        response.headers.get
    """
    if headers is None:
        headers = {}

    response = httpx.Response(
        request=httpx.Request(
            method=method,
            url=httpx.URL(url),
        ),
        status_code=status_code,
        headers=httpx.Headers(headers),
        # Always use ``stream`` rather than content because it's more flexible
        # even if the request is made non-streaming or the response is read with
        # ``response.read()``, httpx will still consume the ``stream`` response
        # content internally. This allows testing both the non-streaming and
        # streaming download paths without needing to complicate the response params
        stream=_IteratorByteSteam(stream),
    )

    response.read = mock.Mock(wraps=response.read)
    response.iter_bytes = mock.Mock(wraps=response.iter_bytes)
    response.headers.get = mock.Mock(wraps=response.headers.get)

    return response


@pytest.fixture
def file_perms() -> int:
    """The expected permissions for the downloaded file.

    Since umask can vary on different systems, it is updated to a known value and reset
    after the test finishes.

    On Windows, the umask seems to always be zero and chmod doesn't really do anything
    anyway.
    """
    if platform.system() == "Windows":
        yield 0o666
    else:
        orig_umask = os.umask(0o077)
        yield 0o600
        os.umask(orig_umask)


@pytest.mark.parametrize(
    ("url", "content_disposition"),
    [
        # A `None` value for `content_disposition` means we skip the header.
        # Other values are passed through as HTTP header values.
        ("https://example.com/path/to/something.zip", None),
        # Ensure empty header is ignored.
        ("https://example.com/path/to/something.zip", ""),
        # Paradigmatic case for Content-Disposition: attachment.
        (
            "https://example.com/path/to/irrelevant.zip",
            "attachment; filename=something.zip",
        ),
        # Ensure extra parameters are ignored.
        (
            "https://example.com/path/to/irrelevant.zip",
            "attachment; filename=something.zip; ignored=okay",
        ),
        # Ensure garbage headers are ignored.
        ("https://example.com/path/to/something.zip", "garbage"),
        # Ensure we respect unusual quoting & case permitted by RFC 6266.
        (
            "https://example.com/path/to/irrelevant.zip",
            'ATTACHment; filename=    "something.zip"',
        ),
        # Ensure we use filename=, even if filename*= is provided. This makes us a
        # "legacy user agent" in the terms of RFC 6266, for our own simplicity.
        (
            "https://example.com/path/to/irrelevant.zip",
            'attachment; filename="something.zip"; filename*=utf-8'
            "''%e2%82%ac%20rates",
        ),
    ],
)
def test_new_download_oneshot(mock_tools, file_perms, url, content_disposition):
    """If no content-length is provided, ``File`` downloadds the file all at once."""
    response = _make_httpx_response(
        method="GET",
        url=url,
        status_code=200,
        headers=(
            {
                "content-disposition": content_disposition,
            }
            if content_disposition is not None
            else {}
        ),
        stream=[b"all content"],
    )
    mock_tools.httpx.stream.return_value.__enter__.return_value = response

    # Download the file
    filename = mock_tools.file.download(
        url="https://example.com/support?useful=Yes",
        download_path=mock_tools.base_path / "downloads",
    )

    # httpx.stream has been invoked, but content isn't iterated
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        "https://example.com/support?useful=Yes",
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )
    response.headers.get.assert_called_with("content-length")
    response.read.assert_called_once()

    # The filename is derived from the URL or header
    assert filename == mock_tools.base_path / "downloads/something.zip"

    # Temporary file was created in download dir and was renamed
    temp_filename = Path(mock_tools.shutil.move.call_args_list[0].args[0])
    assert temp_filename.parent == mock_tools.base_path / "downloads"
    assert temp_filename.name.startswith("something.zip.")
    assert temp_filename.name.endswith(TEMPORARY_DOWNLOAD_FILE_SUFFIX)
    mock_tools.shutil.move.assert_called_with(str(temp_filename), filename)

    # File permissions were set
    mock_tools.os.chmod.assert_called_with(filename, file_perms)
    assert stat.S_IMODE(os.stat(filename).st_mode) == file_perms

    # Attempt to delete Temporary file was made
    mock_tools.os.remove.assert_called_with(str(temp_filename))

    # File content is as expected
    with (mock_tools.base_path / "downloads/something.zip").open(encoding="utf-8") as f:
        assert f.read() == "all content"


def test_new_download_chunked(mock_tools, file_perms):
    """If a content-length is provided, ``File`` streams the response rather than
    downloading it all at once."""
    response = _make_httpx_response(
        method="GET",
        url="https://example.com/path/to/something.zip",
        status_code=200,
        headers={"content-length": "24"},
        stream=[
            b"chunk-1;",
            b"chunk-2;",
            b"chunk-3;",
        ],
    )
    mock_tools.httpx.stream.return_value.__enter__.return_value = response

    # Download the file
    filename = mock_tools.file.download(
        url="https://example.com/support?useful=Yes",
        download_path=mock_tools.base_path,
    )

    # httpx.stream has been invoked, and content is chunked.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        "https://example.com/support?useful=Yes",
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )
    response.headers.get.assert_called_with("content-length")
    response.iter_bytes.assert_called_once_with(chunk_size=1048576)

    # The filename is derived from the URL
    assert filename == mock_tools.base_path / "something.zip"

    # Temporary file was created in download dir and was renamed
    temp_filename = Path(mock_tools.shutil.move.call_args_list[0].args[0])
    assert temp_filename.parent == mock_tools.base_path
    assert temp_filename.name.startswith("something.zip.")
    assert temp_filename.name.endswith(TEMPORARY_DOWNLOAD_FILE_SUFFIX)
    mock_tools.shutil.move.assert_called_with(str(temp_filename), filename)

    # File permissions were set
    mock_tools.os.chmod.assert_called_with(filename, file_perms)
    assert stat.S_IMODE(os.stat(filename).st_mode) == file_perms

    # Attempt to delete Temporary file was made
    mock_tools.os.remove.assert_called_with(str(temp_filename))

    # The downloaded file exists, and content is as expected
    assert filename.exists()
    with (mock_tools.base_path / "something.zip").open(encoding="utf-8") as f:
        assert f.read() == "chunk-1;chunk-2;chunk-3;"


def test_already_downloaded(mock_tools):
    """If the file already exists on disk, it isn't re-downloaded.

    The request is still made to derive the filename, but the content is never streamed.
    """

    # Create an existing file
    existing_file = mock_tools.base_path / "something.zip"
    with existing_file.open("w", encoding="utf-8") as f:
        f.write("existing content")

    url = "https://example.com/path/to/something.zip"

    response = _make_httpx_response(
        status_code=200,
        url=url,
        # Use content and a content-encoding that would cause a DecodeError
        # if ``File`` tried to read the response content.
        # Because the file already exists, ``File`` shouldn't try to read
        # the response, and the ``DecodingError`` won't occur.
        headers={"content-length": "100", "content-encoding": "gzip"},
        stream=[b"definitely not gzip content"],
    )
    mock_tools.httpx.stream.return_value.__enter__.return_value = response

    # Download the file
    filename = mock_tools.file.download(
        url=url,
        download_path=mock_tools.base_path,
    )

    # The GET request will have been made
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        url,
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )

    # The request's Content-Disposition header is consumed to
    # examine the filename; the request is abandoned before
    # any other headers are read.
    response.headers.get.assert_called_once_with("Content-Disposition")

    # but the file existed, so the method returns
    assert filename == existing_file
    assert filename.exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_missing_resource(mock_tools):
    """MissingNetworkResourceError raises for 404 status code."""
    url = "https://example.com/something.zip?useful=Yes"
    response = _make_httpx_response(
        url=url,
        status_code=404,
        stream=[],
    )

    mock_tools.httpx.stream.return_value.__enter__.return_value = response

    # Download the file
    with pytest.raises(MissingNetworkResourceError):
        mock_tools.file.download(
            url=url,
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        url,
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )
    response.headers.get.assert_not_called()

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_bad_resource(mock_tools):
    """BadNetworkResourceError raises for non-200 status code."""
    url = "https://example.com/something.zip?useful=Yes"
    response = _make_httpx_response(
        status_code=500,
        url=url,
        stream=[],
    )

    mock_tools.httpx.stream.return_value.__enter__.return_value = response

    # Download the file
    with pytest.raises(BadNetworkResourceError):
        mock_tools.file.download(
            url=url,
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        url,
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )
    response.headers.get.assert_not_called()

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_iter_bytes_connection_error(mock_tools):
    """NetworkFailure raised if response.iter_bytes() errors and cleans up temporary
    files."""
    url = "https://example.com/something.zip?useful=Yes"
    response = _make_httpx_response(
        status_code=200,
        url=url,
        # Force a real DecodingError by setting the response encoding to
        # gzip with response content that is _not_ gzip
        headers={"content-length": "100", "content-encoding": "gzip"},
        stream=[b"definitely not gzip content"],
    )
    mock_tools.httpx.stream.return_value.__enter__.return_value = response

    # Download the file
    with pytest.raises(NetworkFailure, match=r"Unable to download something\.zip"):
        mock_tools.file.download(
            url="https://example.com/something.zip?useful=Yes",
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        "https://example.com/something.zip?useful=Yes",
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )
    response.headers.get.assert_called_with("content-length")

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not moved
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()

    # Failure happens during response streaming, so the temporary file is create
    # but then also correctly cleaned up after the failure
    temp_filename = Path(mock_tools.os.remove.call_args_list[0].args[0])
    assert temp_filename.parent == mock_tools.base_path
    assert temp_filename.name.startswith("something.zip.")
    assert temp_filename.name.endswith(TEMPORARY_DOWNLOAD_FILE_SUFFIX)
    mock_tools.os.remove.assert_called_with(str(temp_filename))


def test_connection_error(mock_tools):
    """NetworkFailure raised if the connection fails."""
    # Use ftp scheme to force raising a real ProtocolError without needing to mock
    url = "ftp://example.com/something.zip"

    # Use the real httpx for this test instead of the MagicMock'd one from mock_tools
    # Keep using the fixture though, so that it still gets cleaned up after the test
    mock_tools.httpx = mock.Mock(wraps=httpx)

    # Failure leads to filename never being read, so the error message will use the full URL
    # rather than the filename
    with pytest.raises(NetworkFailure, match=f"Unable to download {url}"):
        mock_tools.file.download(
            url=url,
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        url,
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was never created because the failure happens before a response
    # is every received
    assert not (mock_tools.base_path / "something.zip.download").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_redirect_connection_error(mock_tools):
    """NetworkFailure raises if the request leads to too many redirects."""
    mock_tools.httpx.stream.side_effect = [
        httpx.TooManyRedirects("Exceeded max redirects")
    ]

    # Download the file
    with pytest.raises(
        NetworkFailure,
        match=r"Unable to download https\:\/\/example.com\/something\.zip\?useful=Yes",
    ):
        mock_tools.file.download(
            url="https://example.com/something.zip?useful=Yes",
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        "https://example.com/something.zip?useful=Yes",
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_ssl_verification_error(mock_tools):
    """NetworkFailure is raised if the request fails due to SSL."""
    # Mock an SSL verification error
    error = httpx.ConnectError("connection error")
    error.__context__ = httpcore.ConnectError()
    error.__context__.__context__ = ssl.SSLCertVerificationError()
    mock_tools.httpx.stream.side_effect = [error]

    # Download the file
    with pytest.raises(
        NetworkFailure,
        match=(
            r"a connection to the server could not be established "
            r"due to a certificate verification problem"
        ),
    ):
        mock_tools.file.download(
            url="https://example.com/something.zip",
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        "https://example.com/something.zip",
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_unknown_httpcore_connectionerror(mock_tools):
    """NetworkFailure is raised if an unknown core connection error occurs."""
    # Mock a connection error at the level of httpcore
    error = httpx.ConnectError("connection error")
    error.__context__ = httpcore.ConnectError()
    mock_tools.httpx.stream.side_effect = [error]

    # Download the file
    with pytest.raises(
        NetworkFailure,
        match=(
            r"a connection to the server could not be established.\n\n"
            r"The reported cause of the problem "
        ),
    ):
        mock_tools.file.download(
            url="https://example.com/something.zip",
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        "https://example.com/something.zip",
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()


def test_unknown_httpx_connectionerror(mock_tools):
    """NetworkFailure is raised if an unknown httpx connection error occurs."""
    # Mock a connection error at the level of httpx
    error = httpx.ConnectError("connection error")
    mock_tools.httpx.stream.side_effect = [error]

    # Download the file
    with pytest.raises(
        NetworkFailure,
        match=(
            r"a connection to the server could not be established.\n\n"
            r"The reported cause of the problem "
        ),
    ):
        mock_tools.file.download(
            url="https://example.com/something.zip",
            download_path=mock_tools.base_path,
        )

    # httpx.stream has been invoked, but nothing else.
    mock_tools.httpx.stream.assert_called_with(
        "GET",
        "https://example.com/something.zip",
        follow_redirects=True,
        verify=mock_tools.file.ssl_context,
    )

    # The file doesn't exist as a result of the download failure
    assert not (mock_tools.base_path / "something.zip").exists()

    # Temporary file was not created, moved, or deleted
    mock_tools.shutil.move.assert_not_called()
    mock_tools.os.chmod.assert_not_called()
    mock_tools.os.remove.assert_not_called()



import pytest


def test_selection_question(console):
    "A question with a dictionary of options can be presented to the user."
    console.values = ["3"]

    options = {
        "first": "The first option",
        "second": "The second option",
        "third": "The third option",
        "fourth": "The fourth option",
    }
    result = console.selection_question(
        description="Test",
        intro="This is a test",
        options=options,
    )

    # Input is requested once
    assert console.prompts == ["Test: "]

    # "third" is the key of the third option.
    assert result == "third"


def test_selection_question_list(console):
    """If selection_question is given a list of values, they're presented as
    provided."""
    console.values = ["3"]

    options = [
        "The first option",
        "The second option",
        "The third option",
        "The fourth option",
    ]
    result = console.selection_question(
        description="Test",
        intro="This is a test",
        options=options,
    )

    # Input is requested once
    assert console.prompts == ["Test: "]

    # The third option is the third option :-)
    assert result == "The third option"


def test_selection_question_bad_input(console):
    # In order, return:
    #     blank
    #     'asdf'
    #     '10'
    #     '3'
    console.values = ["", "asdf", "10", "3"]
    options = {
        "first": "The first option",
        "second": "The second option",
        "third": "The third option",
        "fourth": "The fourth option",
    }
    result = console.selection_question(
        description="Test",
        intro="This is a test",
        options=options,
    )

    # Input is requested five times; first four cause errors.
    assert console.prompts == ["Test: "] * 4

    # The third option was eventually selected
    assert result == "third"


@pytest.mark.parametrize(("index", "default"), [("1", "first"), ("3", "third")])
def test_selection_question_default(console, index, default):
    """If selection_question has a default, it is returned for no input."""
    # Return an empty response when prompted as though the user press entered
    console.values = [""]

    options = {
        "first": "The first option",
        "second": "The second option",
        "third": "The third option",
        "fourth": "The fourth option",
    }
    result = console.selection_question(
        description="Test",
        intro="This is a test",
        options=options,
        default=default,
    )

    # Input is requested once
    assert console.prompts == [f"Test [{index}]: "]

    # The default option is returned
    assert result == default


def test_override_used(console, capsys):
    """The override is used if valid."""
    override_value = "value"
    assert (
        console.selection_question(
            intro="intro",
            description="My variable",
            default=None,
            options=["value", "some_value", "other_value"],
            override_value="value",
        )
        == "value"
    )
    assert f"Using override value {override_value!r}" in capsys.readouterr().out


def test_override_validation(console, capsys):
    """The override is not used if it is not a valid option."""
    console.values = ["3"]

    result = console.selection_question(
        intro="intro",
        description="My variable",
        default="value",
        options=["value", "some_value", "other_value"],
        override_value="invalid_value",
    )

    output = capsys.readouterr().out

    assert "Invalid override value for My variable: 'invalid_value'" in output

    # "other value" was selected by the user after validation failed.
    assert result == "other_value"


def test_default_value_has_correct_index(console):
    """The default value is used and has the correct index if it is a valid option."""
    console.values = [""]

    result = console.selection_question(
        intro="intro",
        description="My variable",
        default="some_value",
        options=["value", "some_value", "other_value"],
        override_value=None,
    )

    # The result is the default value
    assert result == "some_value"


def test_exception_if_wrong_default(console):
    """An exception is raised if the default value is not a valid option."""
    console.values = [""]

    with pytest.raises(
        ValueError,
        match=r"'invalid_value' is not a valid default value",
    ):
        console.selection_question(
            intro="intro",
            description="My variable",
            default="invalid_value",
            options=["value", "some_value", "other_value"],
            override_value=None,
        )



import pytest

from briefcase.config import validate_url


@pytest.mark.parametrize(
    "url",
    [
        "https://example.com",
    ],
)
def test_valid_url(url):
    """Test that valid URLs are accepted."""
    assert validate_url(url)


@pytest.mark.parametrize(
    "url",
    [
        "not a URL!",  # Free text.
        "file:///usr/local/bin",  # File URL
    ],
)
def test_invalid_url(url):
    """Test that invalid URLs are rejected."""
    with pytest.raises(ValueError, match="Not a valid URL"):
        validate_url(url)


def test_invalid_website_url():
    """Test that non-website URLs are rejected."""
    with pytest.raises(ValueError, match="Not a valid website URL"):
        validate_url("gopher://example.com")  # URL, but not a webpage



import os
from unittest.mock import MagicMock

import pytest

from briefcase.integrations.base import ToolCache
from briefcase.integrations.file import File
from briefcase.integrations.subprocess import Subprocess

WIX_DOWNLOAD_URL = (
    "https://github.com/wixtoolset/wix/releases/download/v5.0.2/wix-cli-x64.msi"
)
WIX_EXE_PATH = "PFiles64/WiX Toolset v5.0/bin/wix.exe"
WIX_UI_PATH = (
    "CFiles64/WixToolset/extensions/WixToolset.UI.wixext/5.0.2/wixext5/"
    "WixToolset.UI.wixext.dll"
)


@pytest.fixture
def mock_tools(tmp_path, mock_tools) -> ToolCache:
    mock_tools.host_os = "Windows"
    mock_tools.os = MagicMock(spec=os)

    # Mock default tools
    mock_tools.subprocess = MagicMock(spec_set=Subprocess)
    mock_tools.file.download = MagicMock(spec_set=File.download)

    return mock_tools


@pytest.fixture
def wix_path(tmp_path):
    return tmp_path / "tools/wix"



import pytest

from briefcase.exceptions import BriefcaseCommandError

from .....utils import create_file


def test_minimal_app_resources(create_command, first_app_templated, tmp_path):
    """Windows apps include, at a minimum, a license as extra resources."""
    # Create a license file
    create_file(tmp_path / "base_path/LICENSE", "This is a license file")

    create_command.install_app_resources(first_app_templated)

    # The extras path exists
    assert (tmp_path / "base_path/build/first-app/windows/app/custom_extras").is_dir()

    # A LICENSE.rtf file has been written.
    assert (tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf").is_file()


def test_post_install_script(create_command, first_app_templated, tmp_path):
    """A post_install script can be provided as extra resources."""
    # Create a license file
    create_file(tmp_path / "base_path/LICENSE", "This is a license file")

    # Create a post-install script
    first_app_templated.post_install_script = "scripts/post_install.bat"
    create_file(tmp_path / "base_path/scripts/post_install.bat", "echo Hello world")

    create_command.install_app_resources(first_app_templated)

    # The post-install script has been written.
    base_path = tmp_path / "base_path/build/first-app/windows/app"
    assert (base_path / "custom_extras/_installer/post_install.bat").is_file()

    # A LICENSE.rtf file has been written.
    assert (tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf").is_file()


def test_non_batch_post_install_script(create_command, first_app_templated, tmp_path):
    """If a post install script isn't a batch file, an error is raised."""
    first_app_templated.post_install_script = "scripts/post_install.sh"
    create_file(tmp_path / "base_path/scripts/post_install.sh", "echo Hello world")

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Windows post-install scripts must be .bat files.",
    ):
        create_command.install_app_resources(first_app_templated)


def test_missing_post_install_script(create_command, first_app_templated):
    """If a post install script is specified, but not present, an error is raised."""
    first_app_templated.post_install_script = "scripts/nonexistent.bat"

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Couldn't find post-install script scripts/nonexistent.bat",
    ):
        create_command.install_app_resources(first_app_templated)


def test_pre_uninstall_script(create_command, first_app_templated, tmp_path):
    """A pre_uninstall script can be provided as extra resources."""
    # Create a license file
    create_file(tmp_path / "base_path/LICENSE", "This is a license file")

    # Create a post-install script
    first_app_templated.pre_uninstall_script = "scripts/pre_uninstall.bat"
    create_file(tmp_path / "base_path/scripts/pre_uninstall.bat", "echo Hello world")

    create_command.install_app_resources(first_app_templated)

    # The pre-uninstall script has been written.
    base_path = tmp_path / "base_path/build/first-app/windows/app"
    assert (base_path / "custom_extras/_installer/pre_uninstall.bat").is_file()

    # A LICENSE.rtf file has been written.
    assert (tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf").is_file()


def test_non_batch_pre_uninstall_script(create_command, first_app_templated, tmp_path):
    """If a post install script isn't a batch file, an error is raised."""
    first_app_templated.pre_uninstall_script = "scripts/pre_uninstall.sh"
    create_file(tmp_path / "base_path/scripts/pre_uninstall.sh", "echo Goodbye world")

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Windows pre-uninstall scripts must be .bat files.",
    ):
        create_command.install_app_resources(first_app_templated)


def test_missing_pre_uninstall_script(create_command, first_app_templated):
    """If a post install script is specified, but not present, an error is raised."""
    first_app_templated.pre_uninstall_script = "scripts/nonexistent.bat"

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Couldn't find pre-uninstall script scripts/nonexistent.bat",
    ):
        create_command.install_app_resources(first_app_templated)


def test_alternate_installer_path(create_command, first_app_templated, tmp_path):
    """Windows installer content can be installed in a custom folder."""
    # Define a custom installer path.
    first_app_templated.installer_path = "install_scripts"

    # Create a license file
    create_file(tmp_path / "base_path/LICENSE", "This is a license file")

    # Create a post-install script
    first_app_templated.post_install_script = "scripts/post_install.bat"
    create_file(tmp_path / "base_path/scripts/post_install.bat", "echo Hello world")

    # Create a pre-uninstall script
    first_app_templated.pre_uninstall_script = "scripts/pre_uninstall.bat"
    create_file(tmp_path / "base_path/scripts/pre_uninstall.bat", "echo Goodbye world")

    create_command.install_app_resources(first_app_templated)

    # The post-install script has been written.
    base_path = tmp_path / "base_path/build/first-app/windows/app"
    assert (base_path / "custom_extras/install_scripts/post_install.bat").is_file()

    # The pre-uninstall script has been written.
    assert (base_path / "custom_extras/install_scripts/post_install.bat").is_file()

    # A LICENSE.rtf file has been written.
    assert (base_path / "LICENSE.rtf").is_file()


def test_legacy_template_support(create_command, first_app_templated, tmp_path):
    """If briefcase.toml doesn't define extras_path; fall back to "extras"."""
    # Create an empty briefcase.toml
    create_file(tmp_path / "base_path/build/first-app/windows/app/briefcase.toml", "")

    # Create a license file
    create_file(tmp_path / "base_path/LICENSE", "This is a license file")

    create_command.install_app_resources(first_app_templated)

    # The extras path exists
    assert (tmp_path / "base_path/build/first-app/windows/app/extras").is_dir()

    # A LICENSE.rtf file has been written.
    assert (tmp_path / "base_path/build/first-app/windows/app/LICENSE.rtf").is_file()



import pytest


@pytest.mark.parametrize(
    ("input", "result"),
    [
        # A GUI app
        (["1"], False),
        # A console app
        (["2"], True),
        # Default value is False (GUI app)
        ([""], False),
        # Invalid values are rejected until a valid value is provided.
        (["x", "y", "2"], True),
    ],
)
def test_app_type(convert_command, input, result):
    """The user can be asked for the app type."""
    convert_command.console.values = input
    out = convert_command.input_app_type(None)
    assert out == result


@pytest.mark.parametrize(
    ("override", "result"),
    [
        # Console is any spelling generates a console app
        ("console", True),
        ("Console", True),
        # Anything else is a GUI app
        ("GUI", False),
        ("gui", False),
        ("Gui", False),
    ],
)
def test_input_app_type_override(convert_command, override, result):
    """The app type is resilient to case changes in the override value."""
    out = convert_command.input_app_type(override)
    assert out == result



import os
import sys

import pytest


@pytest.fixture
def system_path():
    """System PATH for testing."""
    return (
        "C:\\Windows\\system32;C:\\Windows"
        if sys.platform == "win32"
        else "/system/bin"
    )


@pytest.fixture
def user_path():
    """User PATH for testing."""
    return "C:\\custom\\path" if sys.platform == "win32" else "/custom/path"


@pytest.fixture
def complex_path():
    """Complex PATH with multiple entries for testing."""
    return (
        "C:\\system\\bin;C:\\usr\\bin"
        if sys.platform == "win32"
        else "/system/bin:/usr/bin"
    )


@pytest.fixture
def empty_path():
    """Empty PATH for testing."""
    return ""


@pytest.fixture
def venv_bin_dir():
    """Platform-appropriate virtual environment binary directory."""
    return "Scripts" if sys.platform == "win32" else "bin"


@pytest.mark.parametrize("overrides", [None, {}, {"VALUE": "value"}])
def test_full_env_basic_creation(
    venv_context,
    complex_path,
    venv_bin_dir,
    overrides,
    monkeypatch,
):
    """full_env creates an environment with expected defaults."""
    monkeypatch.setenv("PATH", complex_path)

    result = venv_context.full_env(overrides)

    expected_path = (
        str(venv_context.venv_path / venv_bin_dir) + os.pathsep + complex_path
    )
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)

    if overrides and "VALUE" in overrides:
        assert result["VALUE"] == "value"

    assert "PYTHONHOME" not in result


def test_full_env_path_override_custom_path_with_system(
    venv_context,
    user_path,
    system_path,
    venv_bin_dir,
    monkeypatch,
):
    """A path provided as an override supersedes the system path, but is augmented with
    the venv path."""
    overrides = {"PATH": user_path}
    monkeypatch.setenv("PATH", system_path)

    result = venv_context.full_env(overrides)

    expected_path = str(venv_context.venv_path / venv_bin_dir) + os.pathsep + user_path
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)
    assert "PYTHONHOME" not in result


def test_full_env_path_override_none_with_system(
    venv_context,
    system_path,
    venv_bin_dir,
    monkeypatch,
):
    """A None PATH override falls back to system PATH, augmented with venv path."""
    overrides = {"PATH": None}
    monkeypatch.setenv("PATH", system_path)

    result = venv_context.full_env(overrides)

    expected_path = (
        str(venv_context.venv_path / venv_bin_dir) + os.pathsep + system_path
    )
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)
    assert "PYTHONHOME" not in result


def test_full_env_path_override_empty_with_system(
    venv_context,
    empty_path,
    system_path,
    venv_bin_dir,
    monkeypatch,
):
    """An empty PATH override falls back to system PATH, augmented with venv path."""
    overrides = {"PATH": empty_path}
    monkeypatch.setenv("PATH", system_path)

    result = venv_context.full_env(overrides)

    expected_path = (
        str(venv_context.venv_path / venv_bin_dir) + os.pathsep + system_path
    )
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)
    assert "PYTHONHOME" not in result


def test_full_env_path_override_custom_with_empty_system(
    venv_context,
    user_path,
    empty_path,
    venv_bin_dir,
    monkeypatch,
):
    """Custom PATH override is used even when system PATH is empty."""
    overrides = {"PATH": user_path}
    monkeypatch.setenv("PATH", empty_path)

    result = venv_context.full_env(overrides)

    expected_path = str(venv_context.venv_path / venv_bin_dir) + os.pathsep + user_path
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)
    assert "PYTHONHOME" not in result


def test_full_env_path_override_none_with_empty_system(
    venv_context,
    empty_path,
    venv_bin_dir,
    monkeypatch,
):
    """When both override and system PATH are empty, only venv path is used."""
    overrides = {"PATH": None}
    monkeypatch.setenv("PATH", empty_path)

    result = venv_context.full_env(overrides)

    expected_path = str(venv_context.venv_path / venv_bin_dir)
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)
    assert "PYTHONHOME" not in result


def test_full_env_no_overrides(
    venv_context,
    system_path,
    venv_bin_dir,
    monkeypatch,
):
    """full_env with no overrides parameter uses system PATH."""
    monkeypatch.setenv("PATH", system_path)

    result = venv_context.full_env(None)

    expected_path = (
        str(venv_context.venv_path / venv_bin_dir) + os.pathsep + system_path
    )
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)
    assert "PYTHONHOME" not in result


def test_full_env_no_system_path(
    venv_context,
    venv_bin_dir,
    monkeypatch,
):
    """When no system PATH exists, only venv path is used."""
    monkeypatch.delenv("PATH", raising=False)

    result = venv_context.full_env(None)

    expected_path = str(venv_context.venv_path / venv_bin_dir)
    assert result["PATH"] == expected_path
    assert result["VIRTUAL_ENV"] == str(venv_context.venv_path)
    assert "PYTHONHOME" not in result
